from __future__ import annotations

import datetime
import logging
import typing as t
import uuid

from globus_sdk._internal.remarshal import stringify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload

log = logging.getLogger(__name__)


class DeleteData(GlobusPayload):
    r"""
    Convenience class for constructing a delete document, to use as the
    `data` parameter to
    :meth:`submit_delete <globus_sdk.TransferClient.submit_delete>`.

    At least one item must be added using
    :meth:`add_item <globus_sdk.DeleteData.add_item>`.

    :param endpoint: The endpoint ID which is targeted by this deletion Task
    :param label: A string label for the Task
    :param submission_id: A submission ID value fetched via :meth:`get_submission_id
        <globus_sdk.TransferClient.get_submission_id>`. By default, the SDK
        will fetch and populate this field when :meth:`submit_delete
        <globus_sdk.TransferClient.submit_delete>` is called.
    :param recursive: Recursively delete subdirectories on the target endpoint
      [default: ``False``]
    :param ignore_missing: Ignore nonexistent files and directories instead of treating
        them as errors. [default: ``False``]
    :param interpret_globs: Enable expansion of ``\*?[]`` characters in the last
        component of paths, unless they are escaped with a preceding backslash, ``\\``
        [default: ``False``]
    :param deadline: An ISO-8601 timestamp (as a string) or a datetime object which
        defines a deadline for the deletion. At the deadline, even if the data deletion
        is not complete, the job will be canceled. We recommend ensuring that the
        timestamp is in UTC to avoid confusion and ambiguity. Examples of ISO-8601
        timestamps include ``2017-10-12 09:30Z``, ``2017-10-12 12:33:54+00:00``, and
        ``2017-10-12``
    :param notify_on_succeeded: Send a notification email when the delete task
        completes with a status of SUCCEEDED.
        [default: ``True``]
    :param notify_on_failed: Send a notification email when the delete task completes
        with a status of FAILED.
        [default: ``True``]
    :param notify_on_inactive: Send a notification email when the delete task changes
        status to INACTIVE. e.g. From credentials expiring.
        [default: ``True``]
    :param local_user: Optional value passed to identity mapping specifying which local
        user account to map to. Only usable with Globus Connect Server v5 mapped
        collections.
    :param additional_fields: additional fields to be added to the delete
        document. Mostly intended for internal use

    **Examples**

    See the :meth:`submit_delete <globus_sdk.TransferClient.submit_delete>`
    documentation for example usage.

    **External Documentation**

    See the
    `Task document definition
    <https://docs.globus.org/api/transfer/task_submit/#document_types>`_
    and
    `Delete specific fields
    <https://docs.globus.org/api/transfer/task_submit/#delete_specific_fields>`_
    in the REST documentation for more details on Delete Task documents.

    .. automethodlist:: globus_sdk.TransferData
    """

    def __init__(
        self,
        endpoint: uuid.UUID | str,
        *,
        label: str | MissingType = MISSING,
        submission_id: uuid.UUID | str | MissingType = MISSING,
        recursive: bool | MissingType = MISSING,
        ignore_missing: bool | MissingType = MISSING,
        interpret_globs: bool | MissingType = MISSING,
        deadline: str | datetime.datetime | MissingType = MISSING,
        notify_on_succeeded: bool | MissingType = MISSING,
        notify_on_failed: bool | MissingType = MISSING,
        notify_on_inactive: bool | MissingType = MISSING,
        local_user: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = "delete"
        self["DATA"] = []
        self["endpoint"] = endpoint
        self["label"] = label
        self["submission_id"] = submission_id
        self["deadline"] = stringify(deadline)
        self["local_user"] = local_user
        self["recursive"] = recursive
        self["ignore_missing"] = ignore_missing
        self["interpret_globs"] = interpret_globs
        self["notify_on_succeeded"] = notify_on_succeeded
        self["notify_on_failed"] = notify_on_failed
        self["notify_on_inactive"] = notify_on_inactive

        for k, v in self.items():
            log.debug("DeleteData.%s = %s", k, v)

        if additional_fields is not None:
            self.update(additional_fields)
            for option, value in additional_fields.items():
                log.debug(
                    f"DeleteData.{option} = {value} (option passed "
                    "in via additional_fields)"
                )

    def add_item(
        self,
        path: str,
        *,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        """
        Add a file or directory or symlink to be deleted. If any of the paths
        are directories, ``recursive`` must be set True on the top level
        ``DeleteData``. Symlinks will never be followed, only deleted.

        Appends a delete_item document to the DATA key of the delete
        document.

        :param path: Path to the directory or file to be deleted
        :param additional_fields: additional fields to be added to the delete item
        """
        item_data = {
            "DATA_TYPE": "delete_item",
            "path": path,
            **(additional_fields or {}),
        }
        log.debug('DeleteData[{}].add_item: "{}"'.format(self["endpoint"], path))
        self["DATA"].append(item_data)

    def iter_items(self) -> t.Iterator[dict[str, t.Any]]:
        """
        An iterator of items created by ``add_item``.

        Each item takes the form of a dictionary.
        """
        yield from iter(self["DATA"])



import pytest


def test_by_resource_server_lookups(oauth_token_response):
    by_rs = oauth_token_response.by_resource_server

    # things which should hold for all
    for n in (1, 2, 3):
        name = f"resource_server_{n}"
        for attr in ("access_token", "refresh_token", "resource_server"):
            assert by_rs[name][attr] == f"{attr}_{n}"

        assert "expires_in" not in by_rs[name]
        assert "expires_at_seconds" in by_rs[name]

        assert by_rs[name]["token_type"] == "bearer"

    assert by_rs["resource_server_1"]["scope"] == "scope1"
    assert by_rs["resource_server_2"]["scope"] == "scope2 scope2:0 scope2:1"
    assert by_rs["resource_server_3"]["scope"] == "scope3:0 scope3:1"


@pytest.mark.parametrize(
    "scopestr, resource_server",
    [
        ("scope1", "resource_server_1"),
        ("scope2 scope2:0 scope2:1", "resource_server_2"),
        ("scope3:0 scope3:1", "resource_server_3"),
    ],
)
def test_by_scopes_lookups_simple(scopestr, resource_server, oauth_token_response):
    by_scopes = oauth_token_response.by_scopes

    # containment
    assert scopestr in by_scopes
    for x in scopestr.split():
        assert x in by_scopes

    # maps correctly
    assert by_scopes[scopestr]["resource_server"] == resource_server


def test_by_scopes_lookups_failures(oauth_token_response):
    by_scopes = oauth_token_response.by_scopes

    mixed_scopes = "scope1 scope2"

    # containment
    assert mixed_scopes not in by_scopes
    assert "badscope" not in by_scopes

    # try to actually do it, ensure KeyErrors have good messages
    with pytest.raises(KeyError) as excinfo:
        by_scopes[mixed_scopes]
    assert "did not match exactly one token" in str(excinfo.value)

    with pytest.raises(KeyError) as excinfo:
        by_scopes["badscope"]
    assert "was not found" in str(excinfo.value)


def test_by_scopes_lookups_fancy(oauth_token_response):
    by_scopes = oauth_token_response.by_scopes

    # repeated scope
    assert by_scopes["scope3:0 scope3:0"]["resource_server"] == "resource_server_3"

    # not matching order from original document
    assert (
        by_scopes["scope2:1 scope2 scope2:0"]["resource_server"] == "resource_server_2"
    )


def test_stringify(oauth_token_response):
    data = str(oauth_token_response)
    # it starts the right way
    assert data.startswith("OAuthTokenResponse:\n")
    # it contains the id_token, but heavily truncated and annotated as such
    assert "  id_token: id_token_v... (truncated)\n" in data
    # it contains the by_resource_server mapping (but we won't assert anything about
    # that data other than that it starts with a '{')
    assert "  by_resource_server:\n    {" in data


def test_stringify_dependent_tokens(oauth_dependent_token_response):
    data = str(oauth_dependent_token_response)
    # it starts the right way
    assert data.startswith("OAuthDependentTokenResponse:\n")
    # it does not contain the id_token, so this is indicated as None
    assert "  id_token: None\n" in data
    # it contains the by_resource_server mapping (but we won't assert anything about
    # that data other than that it starts with a '{')
    assert "  by_resource_server:\n    {" in data



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="search",
        method="POST",
        path="/v1/index",
        json={
            "@datatype": "GSearchIndex",
            "@version": "2017-09-01",
            "creation_date": "2021-04-05 15:05:18",
            "display_name": "Awesome Index of Awesomeness",
            "description": "An index so awesome that it simply cannot be described",
            "id": INDEX_ID,
            "is_trial": True,
            "subscription_id": None,
            "max_size_in_mb": 1,
            "num_entries": 0,
            "num_subjects": 0,
            "size_in_mb": 0,
            "status": "open",
        },
        metadata={"index_id": INDEX_ID},
    ),
    trial_limit=RegisteredResponse(
        service="search",
        method="POST",
        path="/v1/index",
        status=409,
        json={
            "@datatype": "GError",
            "request_id": "38186e960f3a64c9d530d48ba2271285",
            "status": 409,
            "error_data": {
                "cause": (
                    "When creating an index, an 'owner' role is created "
                    "automatically. If this would exceed ownership limits, this error "
                    "is raised instead."
                ),
                "constraint": (
                    "Cannot create more ownership roles on trial indices "
                    "than the limit (3)"
                ),
            },
            "@version": "2017-09-01",
            "message": "Role limit exceeded",
            "code": "Conflict.LimitExceeded",
        },
    ),
)



import urllib.parse

import pytest

from globus_sdk.testing import get_last_request, load_response


def test_get_group(groups_client):
    meta = load_response(groups_client.get_group).metadata

    res = groups_client.get_group(group_id=meta["group_id"])
    assert res.http_status == 200
    assert "Claptrap" in res["name"]


@pytest.mark.parametrize(
    "include_param",
    ["policies", "policies,memberships", ["memberships", "policies", "child_ids"]],
)
def test_get_group_include(groups_client, include_param):
    meta = load_response(groups_client.get_group).metadata
    expect_param = (
        ",".join(include_param) if not isinstance(include_param, str) else include_param
    )

    res = groups_client.get_group(group_id=meta["group_id"], include=include_param)
    assert res.http_status == 200
    assert "Claptrap" in res["name"]

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    assert len(parsed_qs["include"]) == 1
    assert parsed_qs["include"][0] == expect_param



from .collection import (
    CollectionDocument,
    CollectionPolicies,
    GoogleCloudStorageCollectionPolicies,
    GuestCollectionDocument,
    MappedCollectionDocument,
    POSIXCollectionPolicies,
    POSIXStagingCollectionPolicies,
)
from .endpoint import EndpointDocument
from .role import GCSRoleDocument
from .storage_gateway import (
    ActiveScaleStoragePolicies,
    AzureBlobStoragePolicies,
    BlackPearlStoragePolicies,
    BoxStoragePolicies,
    CephStoragePolicies,
    GoogleCloudStoragePolicies,
    GoogleDriveStoragePolicies,
    HPSSStoragePolicies,
    IrodsStoragePolicies,
    OneDriveStoragePolicies,
    POSIXStagingStoragePolicies,
    POSIXStoragePolicies,
    S3StoragePolicies,
    StorageGatewayDocument,
    StorageGatewayPolicies,
)
from .user_credential import UserCredentialDocument

__all__ = (
    # endpoint documents
    "EndpointDocument",
    # collection documents
    "MappedCollectionDocument",
    "GuestCollectionDocument",
    "CollectionDocument",
    # collection document second-order helpers
    "CollectionPolicies",
    "POSIXCollectionPolicies",
    "POSIXStagingCollectionPolicies",
    "GoogleCloudStorageCollectionPolicies",
    # role document
    "GCSRoleDocument",
    # storage gateway document
    "StorageGatewayDocument",
    # storage gateway document second-order helpers
    "StorageGatewayPolicies",
    "POSIXStoragePolicies",
    "POSIXStagingStoragePolicies",
    "BlackPearlStoragePolicies",
    "BoxStoragePolicies",
    "CephStoragePolicies",
    "GoogleDriveStoragePolicies",
    "GoogleCloudStoragePolicies",
    "OneDriveStoragePolicies",
    "AzureBlobStoragePolicies",
    "S3StoragePolicies",
    "ActiveScaleStoragePolicies",
    "IrodsStoragePolicies",
    "HPSSStoragePolicies",
    # user credential document
    "UserCredentialDocument",
)



import json

from globus_sdk.testing import get_last_request, load_response


def test_set_subscription_admin_verified(groups_client):
    meta = load_response(groups_client.set_subscription_admin_verified).metadata

    res = groups_client.set_subscription_admin_verified(
        group_id=meta["group_id"],
        subscription_id=meta["subscription_id"],
    )
    assert res.http_status == 200
    assert res.data["group_id"] == meta["group_id"]
    assert res.data["subscription_admin_verified_id"] == meta["subscription_id"]

    req = get_last_request()
    req = json.loads(req.body)
    assert req == {"subscription_admin_verified_id": meta["subscription_id"]}



from unittest.mock import Mock, patch

import pytest

from globus_sdk import ConfidentialAppAuthClient, NativeAppAuthClient
from globus_sdk.gare import GlobusAuthorizationParameters
from globus_sdk.login_flows import (
    CommandLineLoginFlowManager,
    LocalServerLoginFlowManager,
)
from globus_sdk.login_flows.command_line_login_flow_manager import (
    CommandLineLoginFlowEOFError,
)
from globus_sdk.testing import load_response


def _mock_input(s):
    print(s)
    return "mock_input"


def test_command_line_login_flower_manager_native(monkeypatch, capsys):
    """
    test CommandLineLoginFlowManager with a NativeAppAuthClient
    """
    login_client = NativeAppAuthClient("mock_client_id")
    load_response(login_client.oauth2_exchange_code_for_tokens)
    monkeypatch.setattr("builtins.input", _mock_input)

    login_flow_manager = CommandLineLoginFlowManager(login_client)
    auth_params = GlobusAuthorizationParameters(
        required_scopes=["urn:globus:auth:scope:transfer.api.globus.org:all"],
        session_required_identities=["user@org.edu"],
    )
    token_res = login_flow_manager.run_login_flow(auth_params)
    assert (
        token_res.by_resource_server["transfer.api.globus.org"]["access_token"]
        == "transfer_access_token"
    )

    captured_output = capsys.readouterr().out
    assert "https://auth.globus.org/v2/oauth2/authorize" in captured_output
    assert "client_id=mock_client_id" in captured_output
    assert "&session_required_identities=user%40org.edu" in captured_output


def test_command_line_login_flower_manager_with_custom_prompts(capsys):
    """
    test CommandLineLoginFlowManager with a NativeAppAuthClient
    """
    login_client = NativeAppAuthClient("mock_client_id")
    load_response(login_client.oauth2_exchange_code_for_tokens)

    class CustomizedLoginFlowManager(CommandLineLoginFlowManager):
        def print_authorize_url(self, authorize_url):
            print("MyCustomLoginPrompt:", authorize_url)

        def prompt_for_code(self):
            return "mock_code_input"

    login_flow_manager = CustomizedLoginFlowManager(login_client)
    auth_params = GlobusAuthorizationParameters(
        required_scopes=["urn:globus:auth:scope:transfer.api.globus.org:all"]
    )
    login_flow_manager.run_login_flow(auth_params)

    captured_output = capsys.readouterr().out
    assert "MyCustomLoginPrompt" in captured_output
    assert "https://auth.globus.org/v2/oauth2/authorize" in captured_output


def test_command_line_login_flower_manager_confidential(monkeypatch, capsys):
    """
    test CommandLineLoginFlowManager with a ConfidentialAppAuthClient
    """
    login_client = ConfidentialAppAuthClient(
        client_id="mock_client_id", client_secret="mock_client_secret"
    )
    load_response(login_client.oauth2_exchange_code_for_tokens)
    monkeypatch.setattr("builtins.input", _mock_input)
    redirect_uri = "https://example.com/callback"

    login_flow_manager = CommandLineLoginFlowManager(
        login_client, redirect_uri=redirect_uri
    )
    auth_params = GlobusAuthorizationParameters(
        required_scopes=["urn:globus:auth:scope:transfer.api.globus.org:all"],
        session_required_single_domain=["org.edu"],
    )
    token_res = login_flow_manager.run_login_flow(auth_params)
    assert (
        token_res.by_resource_server["transfer.api.globus.org"]["access_token"]
        == "transfer_access_token"
    )

    captured_output = capsys.readouterr().out
    assert "Please authenticate with Globus here:" in captured_output
    assert "Enter the resulting Authorization Code here:" in captured_output
    assert "https://auth.globus.org/v2/oauth2/authorize" in captured_output
    assert "client_id=mock_client_id" in captured_output
    assert "&session_required_single_domain=org.edu" in captured_output


def test_command_line_login_flow_manager_eof_error(monkeypatch, capsys):
    """
    test CommandLineLoginFlowManager with a NativeAppAuthClient
    """
    login_client = NativeAppAuthClient("mock_client_id")
    load_response(login_client.oauth2_exchange_code_for_tokens)

    def broken_input(_s):
        raise EOFError()

    monkeypatch.setattr("builtins.input", broken_input)

    login_flow_manager = CommandLineLoginFlowManager(login_client)
    auth_params = GlobusAuthorizationParameters(
        required_scopes=["urn:globus:auth:scope:transfer.api.globus.org:all"],
        session_required_identities=["user@org.edu"],
    )

    with pytest.raises(CommandLineLoginFlowEOFError):
        login_flow_manager.run_login_flow(auth_params)


class MockRedirectServer:
    def __init__(self, *args, **kwargs) -> None:
        self.socket = Mock()
        self.socket.getsockname.return_value = ("", 0)

    def serve_forever(self):
        pass

    def shutdown(self):
        pass

    def wait_for_code(self):
        return "auth_code"


_LOCAL_SERVER_MODULE = (
    "globus_sdk.login_flows.local_server_login_flow_manager."
    "local_server_login_flow_manager"
)


@patch(f"{_LOCAL_SERVER_MODULE}._open_webbrowser", new=lambda url: None)
@patch(f"{_LOCAL_SERVER_MODULE}.RedirectHTTPServer", new=MockRedirectServer)
def test_local_server_login_flower_manager_native():
    """
    test LocalServerLoginManager with a NativeAppAuthClient
    """
    login_client = NativeAppAuthClient("mock_client_id")
    load_response(login_client.oauth2_exchange_code_for_tokens)
    login_flow_manager = LocalServerLoginFlowManager(
        login_client,
    )
    auth_params = GlobusAuthorizationParameters(
        required_scopes=["urn:globus:auth:scope:transfer.api.globus.org:all"],
    )
    token_res = login_flow_manager.run_login_flow(auth_params)
    assert (
        token_res.by_resource_server["transfer.api.globus.org"]["access_token"]
        == "transfer_access_token"
    )


@patch(f"{_LOCAL_SERVER_MODULE}._open_webbrowser", new=lambda url: None)
@patch(f"{_LOCAL_SERVER_MODULE}.RedirectHTTPServer", new=MockRedirectServer)
def test_local_server_login_flower_manager_confidential():
    """
    test LocalServerLoginManager with a ConfidentialAppAuthClient
    """
    login_client = ConfidentialAppAuthClient(
        client_id="mock_client_id", client_secret="mock_client_secret"
    )
    load_response(login_client.oauth2_exchange_code_for_tokens)
    login_flow_manager = LocalServerLoginFlowManager(
        login_client,
    )
    auth_params = GlobusAuthorizationParameters(
        required_scopes=["urn:globus:auth:scope:transfer.api.globus.org:all"],
    )
    token_res = login_flow_manager.run_login_flow(auth_params)
    assert (
        token_res.by_resource_server["transfer.api.globus.org"]["access_token"]
        == "transfer_access_token"
    )



from __future__ import annotations

import typing as t

import globus_sdk

from ..base import TokenStorage
from ..token_data import TokenStorageData
from .context import TokenValidationContext
from .errors import MissingTokenError
from .validators import TokenDataValidator


class ValidatingTokenStorage(TokenStorage):
    """
    A special token storage which provides token data validation hooks.

    See :class:`TokenDataValidator` for details on hook specifics.

    See :class:`TokenStorage` for common interface details.

    :param token_storage: A proxy token storage for this class to pass through store,
        get, and remove requests to.
    :param validators: A collection of validation hooks to call.

    :ivar str | None identity_id: The primary identity ID of the entity which granted
        tokens, if known.
    """

    def __init__(
        self,
        token_storage: TokenStorage,
        *,
        validators: t.Iterable[TokenDataValidator] = (),
    ) -> None:
        self.token_storage = token_storage
        self.validators: list[TokenDataValidator] = list(validators)
        self.identity_id = _identity_id_from_token_data(
            token_storage.get_token_data_by_resource_server()
        )
        super().__init__(namespace=token_storage.namespace)

    def close(self) -> None:
        """Closing a validating storage closes the storage it contains."""
        self.token_storage.close()

    def _make_context(
        self, token_data_by_resource_server: t.Mapping[str, TokenStorageData]
    ) -> TokenValidationContext:
        """
        Build a TokenValidationContext object and potentially update the stored
        ``identity_id`` information in this ValidatingTokenStorage.

        Importantly, this records ``self.identity_id`` into the ``prior_identity_id``
        slot before applying this update.
        """
        context = TokenValidationContext(
            prior_identity_id=self.identity_id,
            token_data_identity_id=_identity_id_from_token_data(
                token_data_by_resource_server
            ),
        )

        if self.identity_id is None:
            self.identity_id = context.token_data_identity_id

        return context

    def store_token_data_by_resource_server(
        self, token_data_by_resource_server: t.Mapping[str, TokenStorageData]
    ) -> None:
        """
        :param token_data_by_resource_server: A dict of TokenStorageData objects
            indexed by their resource server
        """
        context = self._make_context(token_data_by_resource_server)
        for validator in self.validators:
            validator.before_store(token_data_by_resource_server, context)

        self.token_storage.store_token_data_by_resource_server(
            token_data_by_resource_server
        )

    def get_token_data(self, resource_server: str) -> TokenStorageData:
        """
        :param resource_server: A resource server with cached token data.
        :returns: The token data for the given resource server.
        :raises: :exc:`MissingTokenError` if the underlying ``TokenStorage`` does not
            have any token data for the given resource server.
        """
        token_data = self.token_storage.get_token_data(resource_server)
        if token_data is None:
            msg = f"No token data for {resource_server}"
            raise MissingTokenError(msg, resource_server=resource_server)

        token_data_by_resource_server = {token_data.resource_server: token_data}
        context = self._make_context(token_data_by_resource_server)

        for validator in self.validators:
            validator.after_retrieve(token_data_by_resource_server, context)

        return token_data

    def get_token_data_by_resource_server(self) -> dict[str, TokenStorageData]:
        token_data_by_resource_server = (
            self.token_storage.get_token_data_by_resource_server()
        )
        context = self._make_context(token_data_by_resource_server)
        for validator in self.validators:
            validator.after_retrieve(token_data_by_resource_server, context)
        return token_data_by_resource_server

    def remove_token_data(self, resource_server: str) -> bool:
        """
        :param resource_server: The resource server string to remove token data for
        """
        return self.token_storage.remove_token_data(resource_server)

    def _extract_identity_id(
        self, token_response: globus_sdk.OAuthTokenResponse
    ) -> str | None:
        """
        Override determination of the identity_id for a token response.

        When handling a refresh token, use the stored identity ID if possible.
        Otherwise, call the inner token storage's method of lookup.
        """
        if isinstance(token_response, globus_sdk.OAuthRefreshTokenResponse):
            return self.identity_id
        else:
            return self.token_storage._extract_identity_id(token_response)


def _identity_id_from_token_data(
    token_data_by_resource_server: t.Mapping[str, TokenStorageData],
) -> str | None:
    """
    Read token data by resource server and return the ``identity_id`` value
    which was produced.

    :param token_data_by_resource_server: The token data to read for identity_id.

    :raises ValueError: if there is inconsistent ``identity_id`` information
    """
    token_data_identity_ids: set[str] = {
        token_data.identity_id
        for token_data in token_data_by_resource_server.values()
        if token_data.identity_id is not None
    }

    if len(token_data_identity_ids) == 0:
        return None
    elif len(token_data_identity_ids) == 1:
        return token_data_identity_ids.pop()
    else:
        raise ValueError(
            "token_data_by_resource_server contained TokenStorageData objects with "
            f"different identity_id values: {token_data_identity_ids}"
        )



from __future__ import annotations

import datetime
import logging
import typing as t
import uuid

from globus_sdk._internal.remarshal import stringify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload

log = logging.getLogger(__name__)
_sync_level_dict: dict[t.Literal["exists", "size", "mtime", "checksum"], int] = {
    "exists": 0,
    "size": 1,
    "mtime": 2,
    "checksum": 3,
}


def _parse_sync_level(
    sync_level: t.Literal["exists", "size", "mtime", "checksum"] | int | MissingType,
) -> int | MissingType:
    """
    Map sync_level strings to known int values

    Important: if more levels are added in the future you can always pass as an int
    """
    if isinstance(sync_level, str):
        try:
            sync_level = _sync_level_dict[sync_level]
        except KeyError as err:
            raise ValueError(f"Unrecognized sync_level {sync_level}") from err
    return sync_level


class TransferData(GlobusPayload):
    r"""
    Convenience class for constructing a transfer document, to use as the
    ``data`` parameter to
    :meth:`submit_transfer <globus_sdk.TransferClient.submit_transfer>`.

    At least one item must be added using
    :meth:`add_item <globus_sdk.TransferData.add_item>`.

    :param source_endpoint: The endpoint ID of the source endpoint
    :param destination_endpoint: The endpoint ID of the destination endpoint
    :param label: A string label for the Task
    :param submission_id: A submission ID value fetched via :meth:`get_submission_id \
        <globus_sdk.TransferClient.get_submission_id>`. By default, the SDK
        will fetch and populate this field when :meth:`submit_transfer \
        <globus_sdk.TransferClient.submit_transfer>` is called.
    :param sync_level: The method used to compare items between the source and
        destination. One of  ``"exists"``, ``"size"``, ``"mtime"``, or ``"checksum"``
        See the section below on sync-level for an explanation of values.
    :param verify_checksum: When true, after transfer verify that the source and
        destination file checksums match. If they don't, re-transfer the entire file and
        keep trying until it succeeds. This will create CPU load on both the origin and
        destination of the transfer, and may even be a bottleneck if the network speed
        is high enough.
        [default: ``False``]
    :param preserve_timestamp: When true, Globus Transfer will attempt to set file
        timestamps on the destination to match those on the origin. [default: ``False``]
    :param encrypt_data: When true, all files will be TLS-protected during transfer.
        [default: ``False``]
    :param deadline: An ISO-8601 timestamp (as a string) or a datetime object which
        defines a deadline for the transfer. At the deadline, even if the data transfer
        is not complete, the job will be canceled. We recommend ensuring that the
        timestamp is in UTC to avoid confusion and ambiguity. Examples of ISO-8601
        timestamps include ``2017-10-12 09:30Z``, ``2017-10-12 12:33:54+00:00``, and
        ``2017-10-12``
    :param skip_source_errors: When true, source permission denied and file
        not found errors from the source endpoint will cause the offending
        path to be skipped.
        [default: ``False``]
    :param fail_on_quota_errors: When true, quota exceeded errors will cause the
        task to fail.
        [default: ``False``]
    :param delete_destination_extra: Delete files and directories on the
        destination endpoint which don’t exist on the source endpoint or are a
        different type. Only applies for recursive directory transfers.
        [default: ``False``]
    :param notify_on_succeeded: Send a notification email when the transfer completes
        with a status of SUCCEEDED.
        [default: ``True``]
    :param notify_on_failed: Send a notification email when the transfer completes
        with a status of FAILED.
        [default: ``True``]
    :param notify_on_inactive: Send a notification email when the transfer changes
        status to INACTIVE. e.g. From credentials expiring.
        [default: ``True``]
    :param source_local_user: Optional value passed to the source's identity mapping
        specifying which local user account to map to. Only usable with Globus Connect
        Server v5 mapped collections.
    :param destination_local_user: Optional value passed to the destination's identity
        mapping specifying which local user account to map to. Only usable with Globus
        Connect Server v5 mapped collections.
    :param additional_fields: additional fields to be added to the transfer
        document. Mostly intended for internal use

    **Sync Levels**

    The values for ``sync_level`` are used to determine how comparisons are made between
    files found both on the source and the destination. When files match, no data
    transfer will occur.

    For compatibility, this can be an integer ``0``, ``1``, ``2``, or ``3`` in addition
    to the string values.

    The meanings are as follows:

    =====================   ========
    value                   behavior
    =====================   ========
    ``0``, ``exists``       Determine whether or not to transfer based on file
                            existence. If the destination file is absent, do the
                            transfer.
    ``1``, ``size``         Determine whether or not to transfer based on the size of
                            the file. If destination file size does not match the
                            source, do the transfer.
    ``2``, ``mtime``        Determine whether or not to transfer based on modification
                            times. If source has a newer modified time than the
                            destination, do the transfer.
    ``3``, ``checksum``     Determine whether or not to transfer based on checksums of
                            file contents. If source and destination contents differ, as
                            determined by a checksum of their contents, do the transfer.
    =====================   ========

    **Examples**

    See the
    :meth:`submit_transfer <globus_sdk.TransferClient.submit_transfer>`
    documentation for example usage.

    **External Documentation**

    See the
    `Task document definition \
    <https://docs.globus.org/api/transfer/task_submit/#document_types>`_
    and
    `Transfer specific fields \
    <https://docs.globus.org/api/transfer/task_submit/#transfer_specific_fields>`_
    in the REST documentation for more details on Transfer Task documents.

    .. automethodlist:: globus_sdk.TransferData
    """

    def __init__(
        self,
        source_endpoint: uuid.UUID | str,
        destination_endpoint: uuid.UUID | str,
        *,
        label: str | MissingType = MISSING,
        submission_id: uuid.UUID | str | MissingType = MISSING,
        sync_level: (
            int | t.Literal["exists", "size", "mtime", "checksum"] | MissingType
        ) = MISSING,
        verify_checksum: bool | MissingType = MISSING,
        preserve_timestamp: bool | MissingType = MISSING,
        encrypt_data: bool | MissingType = MISSING,
        deadline: datetime.datetime | str | MissingType = MISSING,
        skip_source_errors: bool | MissingType = MISSING,
        fail_on_quota_errors: bool | MissingType = MISSING,
        delete_destination_extra: bool | MissingType = MISSING,
        notify_on_succeeded: bool | MissingType = MISSING,
        notify_on_failed: bool | MissingType = MISSING,
        notify_on_inactive: bool | MissingType = MISSING,
        source_local_user: str | MissingType = MISSING,
        destination_local_user: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        log.debug("Creating a new TransferData object")
        self["DATA_TYPE"] = "transfer"
        self["DATA"] = []
        self["source_endpoint"] = source_endpoint
        self["destination_endpoint"] = destination_endpoint
        self["label"] = label
        self["submission_id"] = submission_id
        self["deadline"] = stringify(deadline)
        self["source_local_user"] = source_local_user
        self["destination_local_user"] = destination_local_user
        self["verify_checksum"] = verify_checksum
        self["preserve_timestamp"] = preserve_timestamp
        self["encrypt_data"] = encrypt_data
        self["skip_source_errors"] = skip_source_errors
        self["fail_on_quota_errors"] = fail_on_quota_errors
        self["delete_destination_extra"] = delete_destination_extra
        self["notify_on_succeeded"] = notify_on_succeeded
        self["notify_on_failed"] = notify_on_failed
        self["notify_on_inactive"] = notify_on_inactive
        self["sync_level"] = _parse_sync_level(sync_level)

        for k, v in self.items():
            log.debug("TransferData.%s = %s", k, v)

        if additional_fields is not None:
            self.update(additional_fields)
            for option, value in additional_fields.items():
                log.debug(
                    f"TransferData.{option} = {value} (option passed "
                    "in via additional_fields)"
                )

    def add_item(
        self,
        source_path: str,
        destination_path: str,
        *,
        recursive: bool | MissingType = MISSING,
        external_checksum: str | MissingType = MISSING,
        checksum_algorithm: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        """
        Add a file or directory to be transferred.

        Appends a transfer_item document to the DATA key of the transfer
        document.

        .. note::

            The full path to the destination file must be provided for file items.
            Parent directories of files are not allowed. See
            `task submission documentation
            <https://docs.globus.org/api/transfer/task_submit/#submit_transfer_task>`_
            for more details.

        :param source_path: Path to the source directory or file to be transferred
        :param destination_path: Path to the destination directory or file will be
            transferred to
        :param recursive: Set to True if the target at source path is a directory
        :param external_checksum: A checksum to verify both source file and destination
            file integrity. The checksum will be verified after the data transfer and a
            failure will cause the entire task to fail. Cannot be used with directories.
            Assumed to be an MD5 checksum unless checksum_algorithm is also given.
        :param checksum_algorithm: Specifies the checksum algorithm to be used when
            verify_checksum is True, sync_level is "checksum" or 3, or an
            external_checksum is given.
        :param additional_fields: additional fields to be added to the transfer item
        """
        item_data: dict[str, t.Any] = {
            "DATA_TYPE": "transfer_item",
            "source_path": source_path,
            "destination_path": destination_path,
            "recursive": recursive,
            "external_checksum": external_checksum,
            "checksum_algorithm": checksum_algorithm,
            **(additional_fields or {}),
        }
        log.debug(
            'TransferData[{}, {}].add_item: "{}"->"{}"'.format(
                self["source_endpoint"],
                self["destination_endpoint"],
                source_path,
                destination_path,
            )
        )
        self["DATA"].append(item_data)

    def add_filter_rule(
        self,
        name: str,
        *,
        method: t.Literal["include", "exclude"] = "exclude",
        type: (  # pylint: disable=redefined-builtin
            t.Literal["file", "dir"] | MissingType
        ) = MISSING,
    ) -> None:
        """
        Add a filter rule to the transfer document.

        These rules specify which items are or are not included when recursively
        transferring directories. Each item that is found during recursive directory
        traversal is matched against these rules in the order they are listed.
        The method of the first filter rule that matches an item is applied (either
        "include" or "exclude"), and filter rule matching stops. If no rules match,
        the item is included in the transfer. Notably, this makes "include" filter
        rules only useful when overriding more general "exclude" filter rules later
        in the list.

        :param name: A pattern to match against item names. Wildcards are supported, as
            are character groups: ``*`` matches everything, ``?`` matches any single
            character, ``[]`` matches any single character within the brackets, and
            ``[!]`` matches any single character not within the brackets.
        :param method: The method to use for filtering. If "exclude" (the default)
            items matching this rule will not be included in the transfer. If
            "include" items matching this rule will be included in the transfer.
        :param type: The types of items on which to apply this filter rule. Either
            ``"file"`` or ``"dir"``. If unspecified, the rule applies to both.
            Note that if a ``"dir"`` is excluded then all items within it will
            also be excluded regardless if they would have matched any include rules.

        Example Usage:

        >>> tdata = TransferData(...)
        >>> tdata.add_filter_rule(method="exclude", "*.tgz", type="file")
        >>> tdata.add_filter_rule(method="exclude", "*.tar.gz", type="file")

        ``tdata`` now describes a transfer which will skip any gzipped tar files with
        the extensions ``.tgz`` or ``.tar.gz``

        >>> tdata = TransferData(...)
        >>> tdata.add_filter_rule(method="include", "*.txt", type="file")
        >>> tdata.add_filter_rule(method="exclude", "*", type="file")

        ``tdata`` now describes a transfer which will only transfer files
        with the ``.txt`` extension.
        """
        if self.get("filter_rules", MISSING) is MISSING:
            self["filter_rules"] = []
        rule = {
            "DATA_TYPE": "filter_rule",
            "method": method,
            "name": name,
            "type": type,
        }
        self["filter_rules"].append(rule)

    def iter_items(self) -> t.Iterator[dict[str, t.Any]]:
        """
        An iterator of items created by ``add_item``.

        Each item takes the form of a dictionary.
        """
        yield from iter(self["DATA"])



from __future__ import annotations

import uuid

import pytest

from globus_sdk import GlobusSDKUsageError
from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "uuid_type",
    (str, uuid.UUID),
)
def test_get_client_by_id(
    service_client,
    uuid_type: type[str] | type[uuid.UUID],
):
    meta = load_response(service_client.get_client).metadata

    if uuid_type is str:
        res = service_client.get_client(client_id=meta["client_id"])
    else:
        res = service_client.get_client(client_id=uuid.UUID(meta["client_id"]))

    assert res["client"]["id"] == meta["client_id"]


def test_get_client_by_fqdn(
    service_client,
):
    meta = load_response(service_client.get_client, case="fqdn").metadata
    res = service_client.get_client(fqdn=meta["fqdn"])

    assert res["client"]["id"] == meta["client_id"]


def test_get_client_exactly_one_of_id_or_fqdn(
    service_client,
):
    with pytest.raises(GlobusSDKUsageError):
        service_client.get_client()

    with pytest.raises(GlobusSDKUsageError):
        service_client.get_client(
            client_id="1b72b72e-5251-454d-af67-0be35911d174",
            fqdn="globus.org",
        )



from globus_sdk.transport import RequestsTransport


# customize the status code tuples
# make sure mypy does not reject these changes for updating the tuple sizes
class CustomRetryStatusTransport(RequestsTransport):
    RETRY_AFTER_STATUS_CODES = (503,)
    TRANSIENT_ERROR_STATUS_CODES = (500,)
    EXPIRED_AUTHORIZATION_STATUS_CODES = (401, 403)



import json
import uuid

import pytest

from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize(
    "admin_id_style", ("string", "list", "set", "uuid", "uuid_list")
)
def test_create_project_admin_id_styles(service_client, admin_id_style):
    meta = load_response(service_client.create_project).metadata

    if admin_id_style == "string":
        admin_ids = meta["admin_id"]
    elif admin_id_style == "list":
        admin_ids = [meta["admin_id"]]
    elif admin_id_style == "set":
        admin_ids = {meta["admin_id"]}
    elif admin_id_style == "uuid":
        admin_ids = uuid.UUID(meta["admin_id"])
    elif admin_id_style == "uuid_list":
        admin_ids = [uuid.UUID(meta["admin_id"])]
    else:
        raise NotImplementedError(f"unknown admin_id_style {admin_id_style}")

    res = service_client.create_project(
        "My Project", "support@globus.org", admin_ids=admin_ids
    )

    assert res["project"]["id"] == meta["id"]

    last_req = get_last_request()
    data = json.loads(last_req.body)
    assert list(data) == ["project"], data  # 'project' is the only key
    assert data["project"]["display_name"] == "My Project", data
    assert data["project"]["contact_email"] == "support@globus.org", data
    assert data["project"]["admin_ids"] == [meta["admin_id"]], data
    assert "admin_group_ids" not in data["project"], data


@pytest.mark.parametrize(
    "group_id_style", ("string", "list", "set", "uuid", "uuid_list")
)
def test_create_project_group_id_styles(service_client, group_id_style):
    meta = load_response(service_client.create_project, case="admin_group").metadata

    if group_id_style == "string":
        group_ids = meta["admin_group_id"]
    elif group_id_style == "list":
        group_ids = [meta["admin_group_id"]]
    elif group_id_style == "set":
        group_ids = {meta["admin_group_id"]}
    elif group_id_style == "uuid":
        group_ids = uuid.UUID(meta["admin_group_id"])
    elif group_id_style == "uuid_list":
        group_ids = [uuid.UUID(meta["admin_group_id"])]
    else:
        raise NotImplementedError(f"unknown group_id_style {group_id_style}")

    res = service_client.create_project(
        "My Project", "support@globus.org", admin_group_ids=group_ids
    )

    assert res["project"]["id"] == meta["id"]

    last_req = get_last_request()
    data = json.loads(last_req.body)
    assert list(data) == ["project"], data  # 'project' is the only key
    assert data["project"]["display_name"] == "My Project", data
    assert data["project"]["contact_email"] == "support@globus.org", data
    assert data["project"]["admin_group_ids"] == [meta["admin_group_id"]], data
    assert "admin_ids" not in data["project"], data



import uuid

from responses.matchers import query_param_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

SCOPE1 = {
    "scope_string": "https://auth.globus.org/scopes/3f33d83f-ec0a-4190-887d-0622e7c4ee9a/manage",  # noqa: E501
    "allows_refresh_token": False,
    "id": str(uuid.uuid1()),
    "advertised": False,
    "required_domains": [],
    "name": "Client manage scope",
    "description": "Manage configuration of this client",
    "client": "3f33d83f-ec0a-4190-887d-0622e7c4ee9a",
    "dependent_scopes": [],
}

SCOPE2 = {
    "scope_string": "https://auth.globus.org/scopes/dfc9a6d3-3373-4a6d-b0a1-b7026d1559d6/view",  # noqa: E501
    "allows_refresh_token": False,
    "id": str(uuid.uuid1()),
    "advertised": False,
    "required_domains": [],
    "name": "Client view scope",
    "description": "View configuration of this client",
    "client": "dfc9a6d3-3373-4a6d-b0a1-b7026d1559d6",
    "dependent_scopes": [],
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path="/v2/api/scopes",
        json={"scopes": [SCOPE1, SCOPE2]},
        metadata={
            "scope_ids": [SCOPE1["id"], SCOPE2["id"]],
        },
    ),
    id=RegisteredResponse(
        service="auth",
        path="/v2/api/scopes",
        json={"scopes": [SCOPE1]},
        match=[query_param_matcher(params={"ids": SCOPE1["id"]})],
        metadata={
            "scope_id": SCOPE1["id"],
        },
    ),
    string=RegisteredResponse(
        service="auth",
        path="/v2/api/scopes",
        json={"scopes": [SCOPE2]},
        match=[query_param_matcher(params={"scope_strings": SCOPE2["scope_string"]})],
        metadata={
            "scope_string": SCOPE2["scope_string"],
        },
    ),
)



import json
import re
import types

import pytest

import globus_sdk


def test_enumerate_fixtures_rejects_wrong_object_type(sphinx_runner, capsys):
    sphinx_runner.ensure_failure(
        ".. enumeratetestingfixtures:: globus_sdk.NullAuthorizer",
    )

    captured = capsys.readouterr()
    err_lines = captured.err.splitlines()

    test_line = None
    for line in err_lines:
        if "BaseClient" in line:
            test_line = line
            break
    else:
        pytest.fail("Didn't find 'BaseClient' in stderr")

    assert re.search(
        r"Expected <class '(\w+\.)+NullAuthorizer'> to be a subclass of BaseClient",
        test_line,
    )


# choose an arbitrary client to test against
def test_enumerate_fixtures_of_search_client(sphinx_runner):
    etree = sphinx_runner.to_etree(
        ".. enumeratetestingfixtures:: globus_sdk.SearchClient",
    )
    assert etree.tag == "document"

    title = etree.find("./section/title")
    assert title.text == "globus_sdk.SearchClient"

    dropdowns = etree.findall("./section/container[@design_component='dropdown']")
    # we don't care about exactly what methods are found and produced as dropdowns
    # we just want to make sure there are "some"
    assert len(dropdowns) > 1

    # grab the first dropdown and inspect it
    first_dropdown = dropdowns[0]
    # find the title, make sure it matches a real method
    fixture_title = first_dropdown.find("./rubric/literal")
    assert fixture_title is not None
    first_method_name = fixture_title.text
    assert hasattr(globus_sdk.SearchClient, first_method_name)
    first_method = getattr(globus_sdk.SearchClient, first_method_name)
    assert isinstance(first_method, types.FunctionType)

    # for each dropdown, there should be a content area and it should contain valid JSON
    for dropdown in dropdowns:
        fixture_title = dropdown.find("./rubric/literal").text
        example_block = dropdown.find("./literal_block")
        assert example_block is not None
        assert example_block.get("language") == "json"
        content = example_block.text
        try:
            json.loads(content)
        except json.JSONDecodeError:
            pytest.fail(
                f"{fixture_title} in SearchClient fixture docs didn't have JSON content"
            )



import pytest

import globus_sdk
from globus_sdk.testing import load_response


# TODO: add data for the success case and test it
@pytest.mark.xfail
def test_userinfo():
    raise NotImplementedError


@pytest.mark.parametrize("casename", ("unauthorized", "forbidden"))
def test_userinfo_error_handling(service_client, casename):
    meta = load_response(service_client.userinfo, case=casename).metadata

    with pytest.raises(globus_sdk.AuthAPIError) as excinfo:
        service_client.userinfo()

    err = excinfo.value
    assert err.http_status == meta["http_status"]
    assert err.code == meta["code"]
    assert err.request_id == meta["error_id"]



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

TUNNEL_ID = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="DELETE",
        path=f"/v2/tunnels/{TUNNEL_ID}",
        json={"data": None, "meta": {"request_id": "ofayi2B4R"}},
        metadata={
            "tunnel_id": TUNNEL_ID,
        },
    ),
)



from __future__ import annotations

import hashlib
import platform


def sha256_string(s: str) -> str:
    return hashlib.sha256(s.encode("utf-8")).hexdigest()


def get_nice_hostname() -> str | None:
    """
    Get the current hostname, with the following added behavior:

    - if it ends in '.local', strip that suffix, as this is a frequent macOS behavior
      'DereksCoolMacbook.local' -> 'DereksCoolMacbook'

    - if the hostname is undiscoverable, return None
    """
    name = platform.node()
    if name.endswith(".local"):
        return name[: -len(".local")]
    return name or None


def slash_join(a: str, b: str | None) -> str:
    """
    Join a and b with a single slash, regardless of whether they already
    contain a trailing/leading slash or neither.

    :param a: the first path component
    :param b: the second path component
    """
    if not b:  # "" or None, don't append a slash
        return a
    if a.endswith("/"):
        if b.startswith("/"):
            return a[:-1] + b
        return a + b
    if b.startswith("/"):
        return a + b
    return a + "/" + b



from __future__ import annotations

import os
import threading
import typing as t
import webbrowser
from contextlib import contextmanager
from string import Template

import globus_sdk
from globus_sdk._internal.utils import get_nice_hostname
from globus_sdk.gare import GlobusAuthorizationParameters
from globus_sdk.login_flows.login_flow_manager import LoginFlowManager

from .errors import LocalServerEnvironmentalLoginError, LocalServerLoginError
from .local_server import DEFAULT_HTML_TEMPLATE, RedirectHandler, RedirectHTTPServer

if t.TYPE_CHECKING:
    from globus_sdk.globus_app import GlobusAppConfig

# a list of text-only browsers which are not allowed for use because they don't work
# with Globus Auth login flows and seize control of the terminal
#
# see the webbrowser library for a list of names:
#   https://github.com/python/cpython/blob/69cdeeb93e0830004a495ed854022425b93b3f3e/Lib/webbrowser.py#L489-L502
BROWSER_DENY_LIST = ["lynx", "www-browser", "links", "elinks", "w3m"]


def _check_remote_session() -> None:
    """
    Try to check if this is being run during a remote session, if so
    raise LocalServerLoginFlowError
    """
    if bool(os.environ.get("SSH_TTY", os.environ.get("SSH_CONNECTION"))):
        raise LocalServerEnvironmentalLoginError(
            "Cannot use LocalServerLoginFlowManager in a remote session"
        )


def _open_webbrowser(url: str) -> None:
    """
    Get a default browser and open given url
    If browser is known to be text-only, or opening fails, raise
    LocalServerLoginFlowError
    """
    try:
        browser = webbrowser.get()
        if hasattr(browser, "name"):
            browser_name = browser.name
        elif hasattr(browser, "_name"):
            # MacOSXOSAScript only supports a public name attribute in py311 and later.
            # https://github.com/python/cpython/issues/82828
            browser_name = browser._name
        else:
            raise LocalServerEnvironmentalLoginError(
                "Unable to determine local browser name."
            )

        if browser_name in BROWSER_DENY_LIST:
            raise LocalServerEnvironmentalLoginError(
                "Cannot use LocalServerLoginFlowManager with "
                f"text-only browser '{browser_name}'"
            )

        if not browser.open(url, new=1):
            raise LocalServerEnvironmentalLoginError(
                f"Failed to open browser '{browser_name}'"
            )
    except webbrowser.Error as exc:
        raise LocalServerEnvironmentalLoginError("Failed to open browser") from exc


class LocalServerLoginFlowManager(LoginFlowManager):
    """
    A login flow manager which uses a locally hosted server to drive authentication-code
    token grants. The local server is used as the authorization redirect URI,
    automatically receiving the auth code from Globus Auth after authentication/consent.

    :param AuthLoginClient login_client: The client that will be making Globus
        Auth API calls required for a login flow.
    :param bool request_refresh_tokens: A signal of whether refresh tokens are expected
        to be requested, in addition to access tokens.
    :param str native_prefill_named_grant: A string to prefill in a Native App login
        flow. This value is only used if the `login_client` is a native client.
    :param Template html_template: Optional HTML Template to be populated with the
        values login_result and post_login_message and displayed to the user. A simple
        default is supplied if not provided which informs the user that the login was
        successful and that they may close the browser window.
    :param tuple[str, int] server_address: Optional tuple of the form (host, port) to
        specify an address to run the local server at. Defaults to ("127.0.0.1", 0).
    """

    def __init__(
        self,
        login_client: globus_sdk.AuthLoginClient,
        *,
        request_refresh_tokens: bool = False,
        native_prefill_named_grant: str | None = None,
        server_address: tuple[str, int] = ("localhost", 0),
        html_template: Template = DEFAULT_HTML_TEMPLATE,
    ) -> None:
        super().__init__(
            login_client,
            request_refresh_tokens=request_refresh_tokens,
            native_prefill_named_grant=native_prefill_named_grant,
        )
        self.server_address = server_address
        self.html_template = html_template

    @classmethod
    def for_globus_app(
        cls,
        app_name: str,
        login_client: globus_sdk.AuthLoginClient,
        config: GlobusAppConfig,
    ) -> LocalServerLoginFlowManager:
        """
        Create a ``LocalServerLoginFlowManager`` for use in a GlobusApp.

        :param app_name: The name of the app. Will be prefilled in native auth flows.
        :param login_client: A client used to make Globus Auth API calls.
        :param config: A GlobusApp-bounded object used to configure login flow manager.
        :raises GlobusSDKUsageError: if app config is incompatible with the manager.
        """
        if config.login_redirect_uri:
            # A "local server" relies on the user being redirected back to the server
            # running on the local machine, so it can't use a custom redirect URI.
            msg = "Cannot define a custom redirect_uri for LocalServerLoginFlowManager."
            raise globus_sdk.GlobusSDKUsageError(msg)
        if not isinstance(login_client, globus_sdk.NativeAppAuthClient):
            # Globus Auth has special provisions for native clients which allow implicit
            # redirect url grant to localhost:<any-port>. This is required for the
            # LocalServerLoginFlowManager to work and is not reproducible in
            # confidential clients.
            msg = "LocalServerLoginFlowManager is only supported for Native Apps."
            raise globus_sdk.GlobusSDKUsageError(msg)

        hostname = get_nice_hostname()
        if hostname:
            prefill = f"{app_name} on {hostname}"
        else:
            prefill = app_name

        return cls(
            login_client,
            request_refresh_tokens=config.request_refresh_tokens,
            native_prefill_named_grant=prefill,
        )

    def run_login_flow(
        self,
        auth_parameters: GlobusAuthorizationParameters,
    ) -> globus_sdk.OAuthTokenResponse:
        """
        Run an interactive login flow using a locally hosted server to get tokens
        for the user.

        :param auth_parameters: ``GlobusAuthorizationParameters`` passed through
            to the authentication flow to control how the user will authenticate.
        :raises LocalServerEnvironmentalLoginError: If the local server login flow
            cannot be run due to known failure conditions such as remote sessions or
            text-only browsers.
        :raises LocalServerLoginError: If the local server login flow fails for any
            reason.
        """
        _check_remote_session()

        with self.background_local_server() as server:
            host, port = server.socket.getsockname()
            redirect_uri = f"http://{host}:{port}"

            # open authorize url in web-browser for user to authenticate
            authorize_url = self._get_authorize_url(auth_parameters, redirect_uri)
            _open_webbrowser(authorize_url)

            # get auth code from server
            auth_code = server.wait_for_code()

        if isinstance(auth_code, BaseException):
            msg = f"Authorization failed with unexpected error:\n{auth_code}."
            raise LocalServerLoginError(msg)

        # get and return tokens
        return self.login_client.oauth2_exchange_code_for_tokens(auth_code)

    @contextmanager
    def background_local_server(self) -> t.Iterator[RedirectHTTPServer]:
        """
        Starts a RedirectHTTPServer in a thread as a context manager.
        """
        server = RedirectHTTPServer(
            server_address=self.server_address,
            handler_class=RedirectHandler,
            html_template=self.html_template,
        )
        thread = threading.Thread(target=server.serve_forever)
        thread.daemon = True
        thread.start()

        yield server

        server.shutdown()



import typing as t
import uuid

from responses.matchers import json_params_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

POLICY_REQUEST_ARGS = {
    "policy_id": str(uuid.uuid1()),
}


def make_request_body(request_args: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    request_body = {}

    for field in [
        "authentication_assurance_timeout",
        "required_mfa",
        "display_name",
        "description",
        "domain_constraints_include",
        "domain_constraints_exclude",
    ]:
        if field in request_args:
            request_body[field] = request_args[field]

    if "project_id" in request_args:
        request_body["project_id"] = str(request_args["project_id"])

    return request_body


def make_response_body(request_args: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    return {
        "project_id": str(request_args.get("project_id", uuid.uuid1())),
        "high_assurance": request_args.get("high_assurance", True),
        "required_mfa": request_args.get("required_mfa", False),
        "authentication_assurance_timeout": request_args.get(
            "authentication_assurance_timeout", 25
        ),
        "display_name": request_args.get(
            "display_name", str(uuid.uuid4()).replace("-", "")
        ),
        "description": request_args.get(
            "description", str(uuid.uuid4()).replace("-", "")
        ),
        "domain_constraints_include": request_args.get("domain_constraints_include"),
        "domain_constraints_exclude": request_args.get("domain_constraints_exclude"),
    }


def register_response(
    args: t.Mapping[str, t.Any],
) -> RegisteredResponse:
    request_args = {**POLICY_REQUEST_ARGS, **args}
    request_body = make_request_body(request_args)
    response_body = make_response_body(request_args)

    return RegisteredResponse(
        service="auth",
        method="PUT",
        path=f"/v2/api/policies/{request_args['policy_id']}",
        json={"policy": response_body},
        metadata={
            # Test functions use 'args' to form request
            "args": request_args,
            # Test functions use 'response' to verify response
            "response": response_body,
        },
        match=[json_params_matcher({"policy": request_body})],
    )


RESPONSES = ResponseSet(
    default=register_response({}),
    project_id_str=register_response({"project_id": str(uuid.uuid1())}),
    project_id_uuid=register_response({"project_id": uuid.uuid1()}),
    authentication_assurance_timeout=register_response(
        {"authentication_assurance_timeout": 9100}
    ),
    required_mfa=register_response({"required_mfa": True}),
    not_required_mfa=register_response({"required_mfa": False}),
    display_name=register_response(
        {"display_name": str(uuid.uuid4()).replace("-", "")}
    ),
    description=register_response({"description": str(uuid.uuid4()).replace("-", "")}),
    no_domain_constrants_include=register_response(
        {"domain_constraints_include": None}
    ),
    empty_domain_constrants_include=register_response(
        {"domain_constraints_include": []}
    ),
    domain_constrants_include=register_response(
        {"domain_constraints_include": ["globus.org", "uchicago.edu"]}
    ),
    no_domain_constrants_exclude=register_response(
        {"domain_constraints_exclude": None}
    ),
    empty_domain_constrants_exclude=register_response(
        {"domain_constraints_exclude": []}
    ),
    domain_constrants_exclude=register_response(
        {"domain_constraints_exclude": ["globus.org", "uchicago.edu"]}
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_FLOW_ID

VALIDATE_RUN_SIMPLE_INPUT_BODY = {"param": "value"}

VALIDATE_INVALID_SIMPLE_INPUT_BODY = {"foo": "bar"}

VALIDATE_RUN_SIMPLE_SUCCESS_RESPONSE = {"message": "validation successful"}

validate_simple_input_request = {
    "body": VALIDATE_RUN_SIMPLE_INPUT_BODY,
}

validate_invalid_simple_input_request = {
    "body": VALIDATE_INVALID_SIMPLE_INPUT_BODY,
}


RESPONSES = ResponseSet(
    metadata={
        "flow_id": TWO_HOP_TRANSFER_FLOW_ID,
        "request_body": VALIDATE_RUN_SIMPLE_INPUT_BODY,
    },
    default=RegisteredResponse(
        service="flows",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}/validate_run",
        method="POST",
        status=200,
        json=VALIDATE_RUN_SIMPLE_SUCCESS_RESPONSE,
    ),
    invalid_input_payload=RegisteredResponse(
        service="flows",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}/validate_run",
        method="POST",
        status=400,
        json={
            "error": {
                "code": "FLOW_INPUT_ERROR",
                "detail": [
                    {
                        "loc": ["$.bool"],
                        "type": "InvalidActionInput",
                        "msg": "'not-a-boolean' is not of type 'boolean'",
                    }
                ],
                "message": "Input failed schema validation with 1 error.",
            },
            "debug_id": "00000000-2572-411b-9fa9-c72fbed2b0bb",
        },
    ),
    invalid_token=RegisteredResponse(
        service="flows",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}/validate_run",
        method="POST",
        status=401,
        json={
            "error": {
                "code": "AUTHENTICATION_ERROR",
                "detail": "Expired or invalid Bearer token",
            },
            "debug_id": "00000000-1ca9-477d-9937-a26c9d9384b9",
        },
    ),
    not_a_flow_starter=RegisteredResponse(
        service="flows",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}/validate_run",
        method="POST",
        status=403,
        json={
            "error": {
                "code": "FORBIDDEN",
                "detail": (
                    "You do not have the necessary permissions to perform this action"
                    " on the flow with id value 00000000-0a9d-4036-a9d7-a77c19515594."
                    " Missing permissions: RUN."
                ),
            },
            "debug_id": "00000000-16c3-4872-8bb5-47db18227cb0",
        },
    ),
)



from .client import GCSClient
from .connector_table import ConnectorTable, GlobusConnectServerConnector
from .data import (
    ActiveScaleStoragePolicies,
    AzureBlobStoragePolicies,
    BlackPearlStoragePolicies,
    BoxStoragePolicies,
    CephStoragePolicies,
    CollectionDocument,
    CollectionPolicies,
    EndpointDocument,
    GCSRoleDocument,
    GoogleCloudStorageCollectionPolicies,
    GoogleCloudStoragePolicies,
    GoogleDriveStoragePolicies,
    GuestCollectionDocument,
    HPSSStoragePolicies,
    IrodsStoragePolicies,
    MappedCollectionDocument,
    OneDriveStoragePolicies,
    POSIXCollectionPolicies,
    POSIXStagingCollectionPolicies,
    POSIXStagingStoragePolicies,
    POSIXStoragePolicies,
    S3StoragePolicies,
    StorageGatewayDocument,
    StorageGatewayPolicies,
    UserCredentialDocument,
)
from .errors import GCSAPIError
from .response import IterableGCSResponse, UnpackingGCSResponse

__all__ = (
    "GCSClient",
    "GCSRoleDocument",
    "EndpointDocument",
    "CollectionDocument",
    "GuestCollectionDocument",
    "MappedCollectionDocument",
    "CollectionPolicies",
    "POSIXCollectionPolicies",
    "POSIXStagingCollectionPolicies",
    "GoogleCloudStorageCollectionPolicies",
    "StorageGatewayDocument",
    "StorageGatewayPolicies",
    "POSIXStoragePolicies",
    "POSIXStagingStoragePolicies",
    "BlackPearlStoragePolicies",
    "BoxStoragePolicies",
    "CephStoragePolicies",
    "GoogleDriveStoragePolicies",
    "GoogleCloudStoragePolicies",
    "OneDriveStoragePolicies",
    "AzureBlobStoragePolicies",
    "S3StoragePolicies",
    "ActiveScaleStoragePolicies",
    "IrodsStoragePolicies",
    "HPSSStoragePolicies",
    "GCSAPIError",
    "IterableGCSResponse",
    "UnpackingGCSResponse",
    "UserCredentialDocument",
    "GlobusConnectServerConnector",
    "ConnectorTable",
)



from .client import (
    AuthClient,
    AuthLoginClient,
    ConfidentialAppAuthClient,
    NativeAppAuthClient,
)
from .data import DependentScopeSpec
from .errors import AuthAPIError
from .flow_managers import (
    GlobusAuthorizationCodeFlowManager,
    GlobusNativeAppFlowManager,
)
from .id_token_decoder import IDTokenDecoder
from .identity_map import IdentityMap
from .response import (
    GetConsentsResponse,
    GetIdentitiesResponse,
    OAuthAuthorizationCodeResponse,
    OAuthClientCredentialsResponse,
    OAuthDependentTokenResponse,
    OAuthRefreshTokenResponse,
    OAuthTokenResponse,
)

__all__ = (
    # client classes
    "AuthClient",
    "AuthLoginClient",
    "NativeAppAuthClient",
    "ConfidentialAppAuthClient",
    # errors
    "AuthAPIError",
    # high-level helpers
    "DependentScopeSpec",
    "IdentityMap",
    "IDTokenDecoder",
    # flow managers
    "GlobusNativeAppFlowManager",
    "GlobusAuthorizationCodeFlowManager",
    # responses
    "GetConsentsResponse",
    "GetIdentitiesResponse",
    "OAuthAuthorizationCodeResponse",
    "OAuthClientCredentialsResponse",
    "OAuthDependentTokenResponse",
    "OAuthRefreshTokenResponse",
    "OAuthTokenResponse",
)



from unittest import mock

import pytest

from globus_sdk.authorizers import ClientCredentialsAuthorizer
from globus_sdk.scopes import Scope

ACCESS_TOKEN = "access_token_1"
EXPIRES_AT = -1
SCOPES = "scopes"


@pytest.fixture
def response():
    r = mock.Mock()
    r.by_resource_server = {
        "rs1": {"expires_at_seconds": -1, "access_token": "access_token_2"}
    }
    return r


@pytest.fixture
def client(response):
    c = mock.Mock()
    c.oauth2_client_credentials_tokens = mock.Mock(return_value=response)
    return c


@pytest.fixture
def authorizer(client):
    return ClientCredentialsAuthorizer(
        client, SCOPES, access_token=ACCESS_TOKEN, expires_at=EXPIRES_AT
    )


def test_get_token_response(authorizer, response, client):
    """
    Calls _get_token_response, confirms that the mock
    ConfidentialAppAuthClient is used and the known data was returned.
    """
    res = authorizer._get_token_response()
    assert res == response
    client.oauth2_client_credentials_tokens.assert_called_once_with(
        requested_scopes=SCOPES
    )


def test_multiple_resource_servers(authorizer, response):
    """
    Sets the mock ConfidentialAppAuthClient to return multiple resource
    servers. Confirms GlobusError is raised when _extract_token_data is
    called.
    """
    response.by_resource_server["rs2"] = {
        "expires_at_seconds": -1,
        "access_token": "access_token_3",
    }
    with pytest.raises(ValueError) as excinfo:
        authorizer._extract_token_data(response)

    assert "didn't return exactly one token" in str(excinfo.value)
    assert SCOPES in str(excinfo.value)


def test_can_create_authorizer_from_scope_objects(client):
    a1 = ClientCredentialsAuthorizer(client, Scope("foo"))
    assert a1.scopes == "foo"

    a2 = ClientCredentialsAuthorizer(
        client, [Scope("foo"), "bar", Scope("baz").with_dependency(Scope("buzz"))]
    )
    assert a2.scopes == "foo bar baz[buzz]"



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

TUNNEL_LIST_DOC = {
    "data": [
        {
            "attributes": {
                "created_time": "2025-12-12T21:11:50.525278",
                "initiator_ip_address": None,
                "initiator_port": None,
                "label": "Buzz Tester",
                "lifetime_mins": 360,
                "listener_ip_address": None,
                "listener_port": None,
                "restartable": False,
                "state": "AWAITING_LISTENER",
                "status": "The tunnel is waiting for listening contact detail setup.",
                "submission_id": "292b0054-7084-46eb-83d6-7a6821b1f77e",
            },
            "id": "1c1be52d-2d4d-4200-b4ad-d75d43eb0d9c",
            "relationships": {
                "initiator": {
                    "data": {
                        "id": "80583f05-75f3-4825-b8a5-6c3edf0bbc5c",
                        "type": "StreamAccessPoint",
                    }
                },
                "listener": {
                    "data": {
                        "id": "dd5fa993-749f-48fb-86cf-f07ad5797d7e",
                        "type": "StreamAccessPoint",
                    }
                },
                "owner": {
                    "data": {
                        "id": "4d443580-012d-4954-816f-e0592bd356e1",
                        "type": "Identity",
                    }
                },
            },
            "type": "Tunnel",
        },
        {
            "attributes": {
                "created_time": "2025-12-12T21:22:11.018233",
                "initiator_ip_address": None,
                "initiator_port": None,
                "label": "part 2",
                "lifetime_mins": 360,
                "listener_ip_address": None,
                "listener_port": None,
                "restartable": False,
                "state": "AWAITING_LISTENER",
                "status": "The tunnel is waiting for listening contact detail setup.",
                "submission_id": "fb3b1220-1d5f-4dcf-92f5-e7056a514319",
            },
            "id": "bf1b0d16-7d93-44eb-8773-9066a750c13e",
            "relationships": {
                "initiator": {
                    "data": {
                        "id": "34c6e671-c011-4bf8-bc30-5ccebada8f3b",
                        "type": "StreamAccessPoint",
                    }
                },
                "listener": {
                    "data": {
                        "id": "dd5fa993-749f-48fb-86cf-f07ad5797d7e",
                        "type": "StreamAccessPoint",
                    }
                },
                "owner": {
                    "data": {
                        "id": "4d443580-012d-4954-816f-e0592bd356e1",
                        "type": "Identity",
                    }
                },
            },
            "type": "Tunnel",
        },
    ],
    "links": None,
    "meta": {"request_id": "fAAfpnino"},
}


RESPONSES = ResponseSet(
    metadata={},
    default=RegisteredResponse(
        service="transfer",
        path="/v2/tunnels",
        json=TUNNEL_LIST_DOC,
        method="GET",
    ),
)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

BOOKMARK_ID = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="DELETE",
        path=f"/v2/bookmarks/{BOOKMARK_ID}",
        status=200,
        json={"meta": {"request_id": "nH45vUMpD"}},
        metadata={
            "bookmark_id": BOOKMARK_ID,
        },
    )
)



import json

from globus_sdk.testing import get_last_request, load_response


def test_batch_delete_by_subject(client):
    meta = load_response(client.batch_delete_by_subject).metadata
    input_subjects = [
        "very-cool-document",
        "less-cool-document",
        "document-wearing-sunglasses",
    ]

    res = client.batch_delete_by_subject(meta["index_id"], subjects=input_subjects)
    assert res.http_status == 200

    assert res["task_id"] == meta["task_id"]

    req = get_last_request()
    sent_data = json.loads(req.body)
    assert sent_data == {"subjects": input_subjects}


def test_batch_delete_by_subject_accepts_string(client):
    """
    Test the handling for a single string.

    We want to ensure that
        subjects="mydoc"
    parses the same as
        subjects=["mydoc"]
    *not* as
        subjects=["m", "y", "d", "o", "c"]
    """
    meta = load_response(client.batch_delete_by_subject).metadata
    input_subject = "very-cool-document"

    res = client.batch_delete_by_subject(meta["index_id"], subjects=input_subject)
    assert res.http_status == 200

    assert res["task_id"] == meta["task_id"]

    req = get_last_request()
    sent_data = json.loads(req.body)
    assert sent_data == {"subjects": [input_subject]}


def test_batch_delete_by_subject_allows_additional_params(client):
    meta = load_response(client.batch_delete_by_subject).metadata
    input_subjects = [
        "very-cool-document",
        "less-cool-document",
        "document-wearing-sunglasses",
    ]

    res = client.batch_delete_by_subject(
        meta["index_id"],
        subjects=input_subjects,
        additional_params={"foo": "snork"},
    )
    assert res.http_status == 200

    assert res["task_id"] == meta["task_id"]

    req = get_last_request()
    sent_data = json.loads(req.body)
    assert sent_data == {"subjects": input_subjects, "foo": "snork"}



import pytest

import globus_sdk


@pytest.fixture
def client():
    # default fqdn for GCS client testing
    client = globus_sdk.GCSClient("abc.xyz.data.globus.org")
    with client.retry_config.tune(max_retries=0):
        yield client



import pytest

import globus_sdk


@pytest.fixture
def compute_client_v2():
    client = globus_sdk.ComputeClientV2()
    with client.retry_config.tune(max_retries=0):
        yield client


@pytest.fixture
def compute_client_v3():
    client = globus_sdk.ComputeClientV3()
    with client.retry_config.tune(max_retries=0):
        yield client



from __future__ import annotations

import typing as t

from docutils import nodes

from .utils import derive_doc_url_base


def extdoclink_role(
    name: str,  # pylint: disable=unused-argument
    rawtext: str,
    text: str,
    lineno: int,  # pylint: disable=unused-argument
    inliner: t.Any,  # pylint: disable=unused-argument
    options: t.Any | None = None,  # pylint: disable=unused-argument
    content: str | None = None,  # pylint: disable=unused-argument
) -> tuple[list[nodes.Node], list[t.Any]]:
    if " " not in text:
        raise ValueError("extdoclink role must contain space-separated text")
    linktext, _, ref = text.rpartition(" ")
    if not (ref.startswith("<") and ref.endswith(">")):
        raise ValueError("extdoclink role reference must be in angle brackets")
    ref = ref[1:-1]
    base_url = derive_doc_url_base(None)
    node = nodes.reference(rawtext, linktext, refuri=f"{base_url}/{ref}")
    return [node], []



from __future__ import annotations

import logging
import typing as t

from .retry import RetryCheck, RetryCheckFlags, RetryCheckResult, RetryContext

log = logging.getLogger(__name__)


class RetryCheckRunner:
    """
    A RetryCheckRunner is an object responsible for running retry checks over the
    lifetime of a request. Unlike the checks or the retry context, the runner persists
    between retries. It can therefore implement special logic for checks like "only try
    this check once".

    Its primary responsibility is to answer the question "should_retry(context)?" with a
    boolean.

    It takes as its input a list of checks. Checks may be paired with flags to indicate
    their configuration options. When not paired with flags, the flags are taken to be
    "NONE".

    Supported flags:

    ``RUN_ONCE``
      The check will run at most once for a given request. Once it has run, it is
      recorded as "has_run" and will not be run again on that request.
    """

    # check configs: a list of pairs, (check, flags)
    # a check without flags is assumed to have flags=NONE
    def __init__(self, checks: t.Iterable[RetryCheck]) -> None:
        self._checks: list[RetryCheck] = []
        self._check_data: dict[RetryCheck, dict[str, t.Any]] = {}
        for check in checks:
            self._checks.append(check)
            self._check_data[check] = {}

    def should_retry(self, context: RetryContext) -> bool:
        for check in self._checks:
            flags = getattr(check, "_retry_check_flags", RetryCheckFlags.NONE)

            if flags & RetryCheckFlags.RUN_ONCE:
                if self._check_data[check].get("has_run"):
                    continue
                else:
                    self._check_data[check]["has_run"] = True

            result = check(context)
            log.debug(  # try to get name but don't fail if it's not a function...
                "ran retry check (%s) => %s", getattr(check, "__name__", check), result
            )
            if result is RetryCheckResult.no_decision:
                continue
            elif result is RetryCheckResult.do_not_retry:
                return False
            else:
                return True

        # fallthrough: don't retry any request which isn't marked for retry
        return False



from globus_sdk.testing import load_response


def test_endpoint_manager_task_successful_transfers(client):
    meta = load_response(client.endpoint_manager_task_successful_transfers).metadata

    response = client.endpoint_manager_task_successful_transfers(meta["task_id"])

    assert response.http_status == 200
    assert response["DATA_TYPE"] == "successful_transfers"



import pytest


def test_extdoclink_role_rejects_no_spaces(sphinxext):
    with pytest.raises(
        ValueError, match="extdoclink role must contain space-separated text"
    ):
        sphinxext.extdoclink_role("ref", "foobar", "foobar", 0, object())


def test_extdoclink_role_requires_angle_brackets(sphinxext):
    with pytest.raises(
        ValueError, match="extdoclink role reference must be in angle brackets"
    ):
        sphinxext.extdoclink_role("ref", "foo bar", "foo bar", 0, object())


def test_extdoclink_role_simple(sphinxext):
    nodes, _ = sphinxext.extdoclink_role(
        "ref", "foo bar &lt;baz/quxx&gt;", "foo bar <baz/quxx>", 0, object()
    )
    assert len(nodes) == 1
    node = nodes[0]
    assert node["refuri"] == "https://docs.globus.org/api/baz/quxx"



from __future__ import annotations

import warnings


class RemovedInV5Warning(DeprecationWarning):
    """
    This warning indicates that a feature or usage was detected which will be
    unsupported in globus-sdk version 5.

    Users are encouraged to resolve these warnings when possible.
    """


def warn_deprecated(message: str, stacklevel: int = 2) -> None:
    warnings.warn(message, RemovedInV5Warning, stacklevel=stacklevel)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import TASK_GROUP_ID, TASK_ID, TASK_ID_2

SUBMIT_RESPONSE = {
    "response": "success",
    "task_group_id": TASK_GROUP_ID,
    "results": [
        {
            "status": "success",
            "task_uuid": TASK_ID,
            "http_status_code": 200,
            "reason": None,
        },
        {
            "status": "success",
            "task_uuid": TASK_ID_2,
            "http_status_code": 200,
            "reason": None,
        },
    ],
}

RESPONSES = ResponseSet(
    metadata={
        "task_group_id": TASK_GROUP_ID,
        "task_id": TASK_ID,
        "task_id_2": TASK_ID_2,
    },
    default=RegisteredResponse(
        service="compute",
        path="/v2/submit",
        method="POST",
        json=SUBMIT_RESPONSE,
    ),
)



import pytest

import globus_sdk
from globus_sdk.testing import get_last_request, load_response


def test_get_identity_providers_by_domains(service_client):
    meta = load_response(service_client.get_identity_providers).metadata
    res = service_client.get_identity_providers(domains=meta["domains"])

    assert [x["id"] for x in res] == meta["ids"]

    lastreq = get_last_request()
    assert lastreq.params == {"domains": ",".join(meta["domains"])}


def test_get_identity_providers_by_ids(service_client):
    meta = load_response(service_client.get_identity_providers).metadata
    res = service_client.get_identity_providers(ids=meta["ids"])

    assert [x["id"] for x in res] == meta["ids"]
    assert [x for y in res for x in y["domains"]] == meta["domains"]

    lastreq = get_last_request()
    assert lastreq.params == {"ids": ",".join(meta["ids"])}


def test_get_identity_providers_mutex_args(service_client):
    with pytest.raises(globus_sdk.GlobusSDKUsageError, match="mutually exclusive"):
        service_client.get_identity_providers(ids="foo", domains="bar")


def test_get_identity_providers_allows_query_params_with_no_args(service_client):
    # this test confirms that the request won't be rejected for passing arguments
    # without specifying either 'ids' or 'domains' -- the supposition being that some
    # other parameter is supported but unknown to the SDK
    meta = load_response(service_client.get_identity_providers).metadata
    res = service_client.get_identity_providers(query_params={"foo": "bar,baz,snork"})

    assert [x["id"] for x in res] == meta["ids"]
    assert [x for y in res for x in y["domains"]] == meta["domains"]

    lastreq = get_last_request()
    assert lastreq.params == {"foo": "bar,baz,snork"}



from __future__ import annotations

import abc
import typing as t
import uuid

from globus_sdk._internal.remarshal import strseq_listify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import AbstractGlobusPayload

from ._common import (
    DatatypeCallback,
    DocumentWithInducedDatatype,
    VersionTuple,
    ensure_datatype,
)

#
# NOTE -- on the organization of arguments in this module --
#
# The arguments to each collection type are defined explicitly for good type annotations
# and documentation.
# However, it's easy for things to get out of sync or different between the various
# locations, so we need to impose some order on things to make comparisons easy.
#
# Complicating this, there are some arguments to specific types which aren't shared
# by the common base.
#
# The rule and rationale used is as follows:
# - DATA_TYPE is special, and always first
# - next, the common optional arguments (shared by all)
# - after that, the specific optional arguments for this type/subtype
# - 'additional_fields' is special, and always last
#
# within those listings of common and specific arguments, the following ordering is
# maintained:
# - strings, sorted alphabetically
# - string lists, sorted alphabetically
# - bools, sorted alphabetically
# - ints, sorted alphabetically
# - dicts and other types, sorted alphabetically
#
# This makes it possible to do side-by-side comparison of common arguments, to ensure
# that they are all present and accounted-for in all contexts, and allows us to compare
# definition lists for param docs and arguments against usage sites to ensure that all
# arguments which are passed are actually used
#


def _user_message_length_callback(
    obj: DocumentWithInducedDatatype,
) -> VersionTuple | None:
    if (
        "user_message" in obj
        and isinstance(obj["user_message"], str)
        and len(obj["user_message"]) > 64
    ):
        return (1, 7, 0)
    return None


# Declare a metaclass of ABCMeta even though inheriting from `Payload` renders it
# inert. This will let type checkers understand that this class is abstract.
class CollectionDocument(AbstractGlobusPayload):
    """
    This is the base class for :class:`~.MappedCollectionDocument` and
    :class:`~.GuestCollectionDocument`.

    Parameters common to both of those are defined and documented here.

    :param data_type: Explicitly set the ``DATA_TYPE`` value for this collection.
        Normally ``DATA_TYPE`` is deduced from the provided parameters and should not be
        set. To maximize compatibility with different versions of GCS, only set this
        value when necessary.

    :param collection_base_path: The location of the collection on its underlying
        storage. For a mapped collection, this is an absolute path on the storage system
        named by the ``storage_gateway_id``. For a guest collection, this is a path
        relative to the value of the ``root_path`` attribute on the mapped collection
        identified by the ``mapped_collection_id``. This parameter is optional for
        updates but required when creating a new collection.
    :param contact_email: Email address of the support contact for the collection
    :param contact_info: Other contact information for the collection, e.g. phone number
        or mailing address
    :param default_directory: Default directory when using the collection
    :param department: The department which operates the collection
    :param description: A text description of the collection
    :param display_name: Friendly name for the collection
    :param identity_id: The Globus Auth identity which acts as the owner of the
        collection
    :param info_link: Link for more info about the collection
    :param organization: The organization which maintains the collection
    :param restrict_transfers_to_high_assurance: Require that transfers of the
        given type involve only collections that are high assurance.
        Valid values: "inbound", "outbound", "all", None
    :param user_message: A message to display to users when interacting with this
        collection
    :param user_message_link: A link to additional messaging for users when interacting
        with this collection

    :param keywords: A list of keywords used to help searches for the collection

    :param disable_verify: Disable verification checksums on transfers to and from this
        collection
    :param enable_https: Enable or disable HTTPS support (requires a managed endpoint)
    :param force_encryption: When set to True, all transfers to and from the collection
        are always encrypted
    :param force_verify: Force verification checksums on transfers to and from this
        collection
    :param public: If True, the collection will be visible to other Globus users

    :param acl_expiration_mins: Length of time that guest collection permissions are
        valid. Only settable on HA guest collections and HA mapped collections and
        used by guest collections attached to it. When set on both the mapped and guest
        collections, the lesser value is in effect.

    :param associated_flow_policy: Policy describing Globus flows to run when the
        collection is accessed. See
        https://docs.globus.org/api/transfer/endpoints_and_collections/#associated_flow_policy
        for expected shape.

    :param additional_fields: Additional data for inclusion in the collection document
    """

    DATATYPE_BASE: str = "collection"
    DATATYPE_VERSION_IMPLICATIONS: dict[str, tuple[int, int, int]] = {
        "associated_flow_policy": (1, 15, 0),
        "activity_notification_policy": (1, 14, 0),
        "auto_delete_timeout": (1, 13, 0),
        "skip_auto_delete": (1, 13, 0),
        "restrict_transfers_to_high_assurance": (1, 12, 0),
        "acl_expiration_mins": (1, 10, 0),
        "delete_protected": (1, 8, 0),
        "guest_auth_policy_id": (1, 6, 0),
        "disable_anonymous_writes": (1, 5, 0),
        "force_verify": (1, 4, 0),
        "sharing_users_allow": (1, 2, 0),
        "sharing_users_deny": (1, 2, 0),
        "enable_https": (1, 1, 0),
        "user_message": (1, 1, 0),
        "user_message_link": (1, 1, 0),
    }
    DATATYPE_VERSION_CALLBACKS: tuple[DatatypeCallback, ...] = (
        _user_message_length_callback,
    )

    def __init__(
        self,
        *,
        # data_type
        data_type: str | MissingType = MISSING,
        # strs
        collection_base_path: str | MissingType = MISSING,
        contact_email: str | None | MissingType = MISSING,
        contact_info: str | None | MissingType = MISSING,
        default_directory: str | MissingType = MISSING,
        department: str | None | MissingType = MISSING,
        description: str | None | MissingType = MISSING,
        display_name: str | MissingType = MISSING,
        identity_id: uuid.UUID | str | MissingType = MISSING,
        info_link: str | None | MissingType = MISSING,
        organization: str | MissingType = MISSING,
        restrict_transfers_to_high_assurance: (
            t.Literal["inbound", "outbound", "all"] | MissingType
        ) = MISSING,
        user_message: str | None | MissingType = MISSING,
        user_message_link: str | None | MissingType = MISSING,
        # str lists
        keywords: t.Iterable[str] | MissingType = MISSING,
        # bools
        disable_verify: bool | MissingType = MISSING,
        enable_https: bool | MissingType = MISSING,
        force_encryption: bool | MissingType = MISSING,
        force_verify: bool | MissingType = MISSING,
        public: bool | MissingType = MISSING,
        # ints
        acl_expiration_mins: int | MissingType = MISSING,
        # dicts
        associated_flow_policy: dict[str, t.Any] | MissingType = MISSING,
        # additional fields
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["collection_type"] = self.collection_type
        self["DATA_TYPE"] = data_type
        self["collection_base_path"] = collection_base_path
        self["contact_email"] = contact_email
        self["contact_info"] = contact_info
        self["default_directory"] = default_directory
        self["department"] = department
        self["description"] = description
        self["display_name"] = display_name
        self["identity_id"] = identity_id
        self["info_link"] = info_link
        self["organization"] = organization
        self["restrict_transfers_to_high_assurance"] = (
            restrict_transfers_to_high_assurance
        )
        self["user_message"] = user_message
        self["user_message_link"] = user_message_link
        self["keywords"] = strseq_listify(keywords)
        self["disable_verify"] = disable_verify
        self["enable_https"] = enable_https
        self["force_encryption"] = force_encryption
        self["force_verify"] = force_verify
        self["public"] = public
        self["acl_expiration_mins"] = acl_expiration_mins
        self["associated_flow_policy"] = associated_flow_policy

        if additional_fields is not MISSING:
            self.update(additional_fields)

    @property
    @abc.abstractmethod
    def collection_type(self) -> str:
        raise NotImplementedError


class MappedCollectionDocument(CollectionDocument):
    """
    An object used to represent a Mapped Collection for creation or update operations.
    The initializer supports all writable fields on Mapped Collections but does not
    include read-only fields like ``id``.

    Because these documents may be used for updates, no fields are strictly required.
    However, GCS will require the following fields for creation:

    - ``storage_gateway_id``
    - ``collection_base_path``

    All parameters for :class:`~.CollectionDocument` are supported in addition to the
    parameters below.

    :param storage_gateway_id: The ID of the storage gateway which hosts this mapped
        collection. This parameter is required when creating a collection.

    :param domain_name: DNS name of the virtual host serving this collection
    :param guest_auth_policy_id: Globus Auth policy ID to set on a mapped collection
        which is then inherited by its guest collections.

    :param sharing_users_allow: Connector-specific usernames allowed to create guest
        collections
    :param sharing_users_deny: Connector-specific usernames forbidden from creating
        guest collections

    :param delete_protected: Enable or disable deletion protection on this collection.
        Defaults to ``True`` during creation.

    :param allow_guest_collections: Enable or disable creation and use of Guest
        Collections on this Mapped Collection
    :param disable_anonymous_writes: Allow anonymous write ACLs on Guest Collections
        attached to this Mapped Collection. This option is only usable on non
        high-assurance collections

    :param auto_delete_timeout: Delete child guest collections that have not been
        accessed within the specified timeout period in days.

    :param policies: Connector-specific collection policies
    :param sharing_restrict_paths: A PathRestrictions document
    """

    @property
    def collection_type(self) -> str:
        return "mapped"

    def __init__(
        self,
        *,
        # data type
        data_type: str | MissingType = MISSING,
        # > common args start <
        # strs
        collection_base_path: str | MissingType = MISSING,
        contact_email: str | None | MissingType = MISSING,
        contact_info: str | None | MissingType = MISSING,
        default_directory: str | MissingType = MISSING,
        department: str | None | MissingType = MISSING,
        description: str | None | MissingType = MISSING,
        display_name: str | MissingType = MISSING,
        identity_id: uuid.UUID | str | MissingType = MISSING,
        info_link: str | None | MissingType = MISSING,
        organization: str | MissingType = MISSING,
        restrict_transfers_to_high_assurance: (
            t.Literal["inbound", "outbound", "all"] | MissingType
        ) = MISSING,
        user_message: str | None | MissingType = MISSING,
        user_message_link: str | None | MissingType = MISSING,
        # str lists
        keywords: t.Iterable[str] | MissingType = MISSING,
        # bools
        disable_verify: bool | MissingType = MISSING,
        enable_https: bool | MissingType = MISSING,
        force_encryption: bool | MissingType = MISSING,
        force_verify: bool | MissingType = MISSING,
        public: bool | MissingType = MISSING,
        # ints
        acl_expiration_mins: int | MissingType = MISSING,
        # > common args end <
        # > specific args start <
        # strs
        domain_name: str | MissingType = MISSING,
        guest_auth_policy_id: uuid.UUID | str | None | MissingType = MISSING,
        storage_gateway_id: uuid.UUID | str | MissingType = MISSING,
        # str lists
        sharing_users_allow: t.Iterable[str] | None | MissingType = MISSING,
        sharing_users_deny: t.Iterable[str] | None | MissingType = MISSING,
        sharing_restrict_paths: dict[str, t.Any] | None | MissingType = MISSING,
        # bools
        delete_protected: bool | MissingType = MISSING,
        allow_guest_collections: bool | MissingType = MISSING,
        disable_anonymous_writes: bool | MissingType = MISSING,
        # ints
        auto_delete_timeout: int | MissingType = MISSING,
        # dicts
        associated_flow_policy: dict[str, t.Any] | MissingType = MISSING,
        policies: CollectionPolicies | dict[str, t.Any] | MissingType = MISSING,
        # > specific args end <
        # additional fields
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__(
            # data type
            data_type=data_type,
            # strings
            collection_base_path=collection_base_path,
            contact_email=contact_email,
            contact_info=contact_info,
            default_directory=default_directory,
            department=department,
            description=description,
            display_name=display_name,
            identity_id=identity_id,
            info_link=info_link,
            organization=organization,
            restrict_transfers_to_high_assurance=restrict_transfers_to_high_assurance,
            user_message=user_message,
            user_message_link=user_message_link,
            # bools
            disable_verify=disable_verify,
            enable_https=enable_https,
            force_encryption=force_encryption,
            force_verify=force_verify,
            public=public,
            # ints
            acl_expiration_mins=acl_expiration_mins,
            # str lists
            keywords=keywords,
            # str dicts
            associated_flow_policy=associated_flow_policy,
            # additional fields
            additional_fields=additional_fields,
        )
        self["domain_name"] = domain_name
        self["restrict_transfers_to_high_assurance"] = (
            restrict_transfers_to_high_assurance
        )
        self["guest_auth_policy_id"] = guest_auth_policy_id
        self["storage_gateway_id"] = storage_gateway_id

        self["sharing_users_allow"] = strseq_listify(sharing_users_allow)
        self["sharing_users_deny"] = strseq_listify(sharing_users_deny)

        self["delete_protected"] = delete_protected
        self["allow_guest_collections"] = allow_guest_collections
        self["disable_anonymous_writes"] = disable_anonymous_writes
        self["auto_delete_timeout"] = auto_delete_timeout
        self["sharing_restrict_paths"] = sharing_restrict_paths
        self["policies"] = policies
        ensure_datatype(self)


class GuestCollectionDocument(CollectionDocument):
    """
    An object used to represent a Guest Collection for creation or update operations.
    The initializer supports all writable fields on Guest Collections but does not
    include read-only fields like ``id``.

    Because these documents may be used for updates, no fields are strictly required.
    However, GCS will require the following fields for creation:

    - ``mapped_collection_id``
    - ``user_credential_id``
    - ``collection_base_path``

    All parameters for :class:`~.CollectionDocument` are supported in addition to the
    parameters below.

    :param mapped_collection_id: The ID of the mapped collection which hosts this guest
        collection
    :param user_credential_id: The ID of the User Credential which is used to access
        data on this collection. This credential must be owned by the collection’s
        ``identity_id``.

    :param skip_auto_delete: Indicates whether the collection is exempt from its
        parent mapped collection's automatic deletion policy.

    :param activity_notification_policy: Specification for when a notification email
        should be sent to a guest collection ``administrator``, ``activity_manager``,
        and ``activity_monitor`` roles when a transfer task reaches completion.
    """

    @property
    def collection_type(self) -> str:
        return "guest"

    def __init__(
        self,
        *,
        # data type
        data_type: str | MissingType = MISSING,
        # > common args start <
        # strs
        collection_base_path: str | MissingType = MISSING,
        contact_email: str | None | MissingType = MISSING,
        contact_info: str | None | MissingType = MISSING,
        default_directory: str | MissingType = MISSING,
        department: str | None | MissingType = MISSING,
        description: str | None | MissingType = MISSING,
        display_name: str | MissingType = MISSING,
        identity_id: uuid.UUID | str | MissingType = MISSING,
        info_link: str | None | MissingType = MISSING,
        organization: str | MissingType = MISSING,
        restrict_transfers_to_high_assurance: (
            t.Literal["inbound", "outbound", "all"] | MissingType
        ) = MISSING,
        user_message: str | None | MissingType = MISSING,
        user_message_link: str | None | MissingType = MISSING,
        # str lists
        keywords: t.Iterable[str] | MissingType = MISSING,
        # bools
        disable_verify: bool | MissingType = MISSING,
        enable_https: bool | MissingType = MISSING,
        force_encryption: bool | MissingType = MISSING,
        force_verify: bool | MissingType = MISSING,
        public: bool | MissingType = MISSING,
        # ints
        acl_expiration_mins: int | MissingType = MISSING,
        # dicts
        associated_flow_policy: dict[str, t.Any] | MissingType = MISSING,
        # > common args end <
        # > specific args start <
        mapped_collection_id: uuid.UUID | str | MissingType = MISSING,
        user_credential_id: uuid.UUID | str | MissingType = MISSING,
        skip_auto_delete: bool | MissingType = MISSING,
        activity_notification_policy: dict[str, list[str]] | MissingType = MISSING,
        # > specific args end <
        # additional fields
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__(
            # data type
            data_type=data_type,
            # strings
            collection_base_path=collection_base_path,
            contact_email=contact_email,
            contact_info=contact_info,
            default_directory=default_directory,
            department=department,
            description=description,
            display_name=display_name,
            identity_id=identity_id,
            info_link=info_link,
            organization=organization,
            restrict_transfers_to_high_assurance=restrict_transfers_to_high_assurance,
            user_message=user_message,
            user_message_link=user_message_link,
            # bools
            disable_verify=disable_verify,
            enable_https=enable_https,
            force_encryption=force_encryption,
            force_verify=force_verify,
            public=public,
            # ints
            acl_expiration_mins=acl_expiration_mins,
            # str lists
            keywords=keywords,
            # dicts
            associated_flow_policy=associated_flow_policy,
            # additional fields
            additional_fields=additional_fields,
        )

        self["mapped_collection_id"] = mapped_collection_id
        self["user_credential_id"] = user_credential_id
        self["skip_auto_delete"] = skip_auto_delete
        self["activity_notification_policy"] = activity_notification_policy
        ensure_datatype(self)


class CollectionPolicies(AbstractGlobusPayload):
    """
    This is the abstract base type for Collection Policies documents to use as the
    ``policies`` parameter when creating a MappedCollectionDocument.
    """


class POSIXCollectionPolicies(CollectionPolicies):
    """
    Convenience class for constructing a Posix Collection Policy
    document to use as the `policies` parameter when creating a
    CollectionDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param sharing_groups_allow: POSIX groups which are allowed to create guest
        collections.
    :param sharing_groups_deny: POSIX groups which are not allowed to create guest
        collections.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "posix_collection_policies#1.0.0",
        sharing_groups_allow: str | t.Iterable[str] | None | MissingType = MISSING,
        sharing_groups_deny: str | t.Iterable[str] | None | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE

        self["sharing_groups_allow"] = strseq_listify(sharing_groups_allow)
        self["sharing_groups_deny"] = strseq_listify(sharing_groups_deny)

        if additional_fields is not MISSING:
            self.update(additional_fields)


class POSIXStagingCollectionPolicies(CollectionPolicies):
    """
    Convenience class for constructing a Posix Staging Collection Policy
    document to use as the ``policies`` parameter when creating a
    CollectionDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param sharing_groups_allow: POSIX groups which are allowed to create guest
        collections.
    :param sharing_groups_deny: POSIX groups which are not allowed to create guest
        collections.
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "posix_staging_collection_policies#1.0.0",
        sharing_groups_allow: str | t.Iterable[str] | None | MissingType = MISSING,
        sharing_groups_deny: str | t.Iterable[str] | None | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["sharing_groups_allow"] = strseq_listify(sharing_groups_allow)
        self["sharing_groups_deny"] = strseq_listify(sharing_groups_deny)

        if additional_fields is not MISSING:
            self.update(additional_fields)


class GoogleCloudStorageCollectionPolicies(CollectionPolicies):
    """
    Convenience class for constructing a Google Cloud Storage Collection Policy
    document to use as the ``policies`` parameter when creating a CollectionDocument

    :param DATA_TYPE: Versioned document type. Defaults to the appropriate type for
        this class.
    :param project: Google Cloud Platform project ID that is used by this collection
    :param additional_fields: Additional data for inclusion in the policy document
    """

    def __init__(
        self,
        DATA_TYPE: str = "google_cloud_storage_collection_policies#1.0.0",
        project: str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["project"] = project
        if additional_fields is not MISSING:
            self.update(additional_fields)



import pytest

from globus_sdk.experimental import TransferClientV2


@pytest.fixture
def client():
    client = TransferClientV2()
    with client.retry_config.tune(max_retries=0):
        yield client



from globus_sdk.testing import get_last_request, load_response


def test_delete_bookmark(client):
    meta = load_response(client.delete_bookmark).metadata

    res = client.delete_bookmark(meta["bookmark_id"])

    assert res.http_status == 200

    req = get_last_request()
    assert req.body is None



import pytest

import globus_sdk
from globus_sdk.testing import RegisteredResponse


def test_transfer_client_will_retry_ordinary_502(client, mocksleep):
    # turn on retries (fixture defaults off)
    client.retry_config.max_retries = 1

    RegisteredResponse(service="transfer", path="/foo", status=502, body="Uh-oh!").add()
    RegisteredResponse(service="transfer", path="/foo", json={"status": "ok"}).add()

    # no sign of an error in the client
    res = client.get("/foo")
    assert res.http_status == 200
    assert res["status"] == "ok"

    # there was a sleep (retry was triggered)
    mocksleep.assert_called_once()


def test_transfer_client_will_not_retry_endpoint_error(client, mocksleep):
    # turn on retries (fixture defaults off)
    client.retry_config.max_retries = 1

    RegisteredResponse(
        service="transfer",
        path="/do_a_gcp_thing",
        status=502,
        json={
            "HTTP status": "502",
            "code": "ExternalError.DirListingFailed.GCDisconnected",
            "error_name": "Transfer API Error",
            "message": "The GCP endpoint is not currently connected to Globus",
            "request_id": "rhvcR0aHX",
        },
    ).add()
    RegisteredResponse(
        service="transfer", path="/do_a_gcp_thing", json={"status": "ok"}
    ).add()

    # no sign of an error in the client
    with pytest.raises(globus_sdk.TransferAPIError) as excinfo:
        client.get("/do_a_gcp_thing")
    assert excinfo.value.http_status == 502

    # there was no sleep (retry was not triggered)
    mocksleep.assert_not_called()



from globus_sdk.testing import get_last_request, load_response


def test_get_bookmark(client):
    meta = load_response(client.get_bookmark).metadata

    res = client.get_bookmark(
        meta["bookmark_id"], query_params={"include": "collection"}
    )

    assert res.http_status == 200
    assert res["data"]["type"] == "Bookmark"
    assert res["data"]["id"] == meta["bookmark_id"]
    assert (
        res["data"]["relationships"]["collection"]["data"]["id"] == meta["collection"]
    )

    req = get_last_request()
    assert "include=collection" in req.url
    assert req.body is None



from __future__ import annotations

import logging
import os
import time
from unittest import mock

import pytest

import globus_sdk
from globus_sdk import (
    AccessTokenAuthorizer,
    AuthLoginClient,
    ClientApp,
    ClientCredentialsAuthorizer,
    ConfidentialAppAuthClient,
    GlobusAppConfig,
    NativeAppAuthClient,
    RefreshTokenAuthorizer,
    TransferClient,
    UserApp,
)
from globus_sdk.exc import GlobusSDKUsageError
from globus_sdk.gare import GlobusAuthorizationParameters
from globus_sdk.globus_app.authorizer_factory import (
    AccessTokenAuthorizerFactory,
    ClientCredentialsAuthorizerFactory,
    RefreshTokenAuthorizerFactory,
)
from globus_sdk.login_flows import (
    CommandLineLoginFlowManager,
    LocalServerLoginFlowManager,
    LoginFlowManager,
)
from globus_sdk.scopes import AuthScopes, Scope
from globus_sdk.testing import load_response
from globus_sdk.token_storage import (
    HasRefreshTokensValidator,
    JSONTokenStorage,
    MemoryTokenStorage,
    NotExpiredValidator,
    SQLiteTokenStorage,
    TokenStorageData,
)


def _mock_token_data_by_rs(
    resource_server: str = "auth.globus.org",
    scope: str = "openid",
    refresh_token: str | None = "mock_refresh_token",
    expiration_delta: int = 300,
):
    return {
        resource_server: TokenStorageData(
            resource_server=resource_server,
            identity_id="mock_identity_id",
            scope=scope,
            access_token="mock_access_token",
            refresh_token=refresh_token,
            expires_at_seconds=int(time.time() + expiration_delta),
            token_type="Bearer",
        )
    }


def _mock_input(s):
    print(s)
    return "mock_input"


def _mock_decode(*args, **kwargs):
    return {"sub": "user_id"}


def test_user_app_native():
    client_id = "mock_client_id"
    user_app = UserApp("test-app", client_id=client_id)

    assert user_app.app_name == "test-app"
    assert isinstance(user_app._login_client, NativeAppAuthClient)
    assert user_app._login_client.app_name == "test-app"
    assert isinstance(user_app._authorizer_factory, AccessTokenAuthorizerFactory)
    assert isinstance(user_app._login_flow_manager, CommandLineLoginFlowManager)


def test_user_app_login_client():
    mock_client = mock.Mock(
        spec=NativeAppAuthClient,
        client_id="mock-client_id",
        base_url="https://auth.globus.org",
        environment="production",
    )
    user_app = UserApp("test-app", login_client=mock_client)

    assert user_app.app_name == "test-app"
    assert user_app._login_client == mock_client
    assert user_app.client_id == "mock-client_id"


def test_user_app_no_client_or_id():
    msg = (
        "Could not set up a globus login client. One of client_id or login_client is "
        "required."
    )
    with pytest.raises(GlobusSDKUsageError, match=msg):
        UserApp("test-app")


def test_user_app_both_client_and_id():
    msg = "Mutually exclusive parameters: client_id and login_client."
    with pytest.raises(GlobusSDKUsageError, match=msg):
        UserApp("test-app", login_client=mock.Mock(), client_id="client_id")


def test_user_app_login_client_environment_mismatch():
    mock_client = mock.Mock(environment="sandbox")

    with pytest.raises(GlobusSDKUsageError) as exc:
        config = GlobusAppConfig(environment="preview")
        UserApp("test-app", login_client=mock_client, config=config)

    expected = "[Environment Mismatch] The login_client's environment (sandbox) does not match the GlobusApp's configured environment (preview)."  # noqa
    assert str(exc.value) == expected


def test_user_app_default_token_storage():
    client_id = "mock_client_id"
    user_app = UserApp("test-app", client_id=client_id)

    token_storage = user_app._authorizer_factory.token_storage.token_storage
    assert isinstance(token_storage, JSONTokenStorage)

    if os.name == "nt":
        # on the windows-latest run this was
        # C:\Users\runneradmin\AppData\Roaming\globus\app\mock_client_id\test-app\tokens.json
        expected = "\\globus\\app\\mock_client_id\\test-app\\tokens.json"
        assert token_storage.filepath.endswith(expected)
    else:
        expected = "~/.globus/app/mock_client_id/test-app/tokens.json"
        assert token_storage.filepath == os.path.expanduser(expected)


class CustomMemoryTokenStorage(MemoryTokenStorage):
    pass


@pytest.mark.parametrize(
    "token_storage_value, token_storage_class",
    (
        # Named token storage types
        ("json", JSONTokenStorage),
        ("sqlite", SQLiteTokenStorage),
        ("memory", MemoryTokenStorage),
        # Custom token storage class (instantiated or class)
        (CustomMemoryTokenStorage(), CustomMemoryTokenStorage),
        (CustomMemoryTokenStorage, CustomMemoryTokenStorage),
    ),
)
def test_user_app_token_storage_configuration(token_storage_value, token_storage_class):
    client_id = "mock_client_id"
    config = GlobusAppConfig(token_storage=token_storage_value)

    user_app = UserApp("test-app", client_id=client_id, config=config)
    if hasattr(user_app._token_storage, "close"):
        user_app._token_storage.close()  # Prevent a ResourceWarning on Python 3.13
    assert isinstance(user_app._token_storage, token_storage_class)


def test_user_app_registers_openid_scope_implicitly():
    client_id = "mock_client_id"
    user_app = UserApp("test-app", client_id=client_id)

    assert "auth.globus.org" in user_app.scope_requirements
    scopes = user_app.scope_requirements["auth.globus.org"]
    assert "openid" in [str(s) for s in scopes]


def test_user_app_with_refresh_tokens_sets_expected_validators():
    client_id = "mock_client_id"
    config = GlobusAppConfig(request_refresh_tokens=True)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    validator_types = {type(x) for x in user_app.token_storage.validators}
    assert HasRefreshTokensValidator in validator_types
    assert NotExpiredValidator not in validator_types


def test_user_app_without_refresh_tokens_sets_expected_validators():
    client_id = "mock_client_id"
    config = GlobusAppConfig(request_refresh_tokens=False)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    validator_types = {type(x) for x in user_app.token_storage.validators}
    assert HasRefreshTokensValidator not in validator_types
    assert NotExpiredValidator in validator_types


class MockLoginFlowManager(LoginFlowManager):
    def __init__(self, login_client: AuthLoginClient | None = None) -> None:
        login_client = login_client or mock.Mock(spec=NativeAppAuthClient)
        super().__init__(login_client)

    @classmethod
    def for_globus_app(
        cls, app_name: str, login_client: AuthLoginClient, config: GlobusAppConfig
    ) -> MockLoginFlowManager:
        return cls(login_client)

    def run_login_flow(self, auth_parameters: GlobusAuthorizationParameters):
        return mock.Mock()


@pytest.mark.parametrize(
    "value,login_flow_manager_class",
    (
        (None, CommandLineLoginFlowManager),
        ("command-line", CommandLineLoginFlowManager),
        ("local-server", LocalServerLoginFlowManager),
        (MockLoginFlowManager(), MockLoginFlowManager),
        (MockLoginFlowManager, MockLoginFlowManager),
    ),
)
def test_user_app_login_flow_manager_configuration(value, login_flow_manager_class):
    client_id = "mock_client_id"
    config = GlobusAppConfig(login_flow_manager=value)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    assert isinstance(user_app._login_flow_manager, login_flow_manager_class)


def test_user_app_templated():
    client_id = "mock_client_id"
    client_secret = "mock_client_secret"
    config = GlobusAppConfig(login_redirect_uri="https://example.com")
    user_app = UserApp(
        "test-app", client_id=client_id, client_secret=client_secret, config=config
    )

    assert user_app.app_name == "test-app"
    assert isinstance(user_app._login_client, ConfidentialAppAuthClient)
    assert user_app._login_client.app_name == "test-app"
    assert isinstance(user_app._authorizer_factory, AccessTokenAuthorizerFactory)
    assert isinstance(user_app._login_flow_manager, CommandLineLoginFlowManager)


def test_user_app_refresh():
    client_id = "mock_client_id"
    config = GlobusAppConfig(request_refresh_tokens=True)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    assert user_app.app_name == "test-app"
    assert isinstance(user_app._login_client, NativeAppAuthClient)
    assert user_app._login_client.app_name == "test-app"
    assert isinstance(user_app._authorizer_factory, RefreshTokenAuthorizerFactory)


def test_client_app():
    client_id = "mock_client_id"
    client_secret = "mock_client_secret"
    client_app = ClientApp("test-app", client_id=client_id, client_secret=client_secret)

    assert client_app.app_name == "test-app"
    assert isinstance(client_app._login_client, ConfidentialAppAuthClient)
    assert client_app._login_client.app_name == "test-app"
    assert isinstance(
        client_app._authorizer_factory, ClientCredentialsAuthorizerFactory
    )


def test_client_app_no_secret():
    client_id = "mock_client_id"

    msg = "A ClientApp requires a client_secret to initialize its own login client"
    with pytest.raises(GlobusSDKUsageError, match=msg):
        ClientApp("test-app", client_id=client_id)


def test_add_scope_requirements_and_auth_params_with_required_scopes():
    client_id = "mock_client_id"
    user_app = UserApp("test-app", client_id=client_id)

    # default without adding requirements is just auth's openid scope
    params = user_app._auth_params_with_required_scopes()
    assert params.required_scopes == ["openid"]

    # re-adding openid alongside other auth scopes, openid shouldn't be duplicated
    user_app.add_scope_requirements(
        {"auth.globus.org": [Scope("openid"), Scope("email"), Scope("profile")]}
    )
    params = user_app._auth_params_with_required_scopes()
    assert sorted(params.required_scopes) == ["email", "openid", "profile"]

    # adding a requirement with a dependency
    user_app.add_scope_requirements(
        {"foo": [Scope("foo:all").with_dependency(Scope("bar:all"))]}
    )
    params = user_app._auth_params_with_required_scopes()
    assert sorted(params.required_scopes) == [
        "email",
        "foo:all[bar:all]",
        "openid",
        "profile",
    ]

    # re-adding a requirement with a new dependency, dependencies should be combined
    user_app.add_scope_requirements(
        {"foo": [Scope("foo:all").with_dependency(Scope("baz:all"))]}
    )
    params = user_app._auth_params_with_required_scopes()
    # order of dependencies is not guaranteed
    assert sorted(params.required_scopes) in (
        ["email", "foo:all[bar:all baz:all]", "openid", "profile"],
        ["email", "foo:all[baz:all bar:all]", "openid", "profile"],
    )


@pytest.mark.parametrize(
    "scope_collection",
    ("email", AuthScopes.email, Scope("email"), [Scope("email")]),
)
def test_add_scope_requirements_accepts_different_scope_types(scope_collection):
    client_id = "mock_client_id"
    user_app = UserApp("test-app", client_id=client_id)

    assert _sorted_auth_scope_str(user_app) == "openid"

    # Add a scope scope string
    user_app.add_scope_requirements({"auth.globus.org": scope_collection})
    assert _sorted_auth_scope_str(user_app) == "email openid"


@pytest.mark.parametrize(
    "scope_collection",
    ("email", AuthScopes.email, Scope("email"), [Scope("email")]),
)
def test_constructor_scope_requirements_accepts_different_scope_types(scope_collection):
    client_id = "mock_client_id"
    user_app = UserApp(
        "test-app",
        client_id=client_id,
        scope_requirements={"auth.globus.org": scope_collection},
    )

    assert _sorted_auth_scope_str(user_app) == "email openid"


def test_scope_requirements_returns_copies_scopes():
    user_app = UserApp("test-app", client_id="mock_client_id")
    foo_scope = Scope("foo:all").with_dependency(Scope("bar:all"))
    user_app.add_scope_requirements({"foo": [foo_scope]})

    real_requirements = user_app._scope_requirements
    real_openid = real_requirements["auth.globus.org"][0]
    real_foo = real_requirements["foo"][0]

    copied_requirements = user_app.scope_requirements
    copied_openid = copied_requirements["auth.globus.org"][0]
    copied_foo = copied_requirements["foo"][0]

    assert real_requirements is not copied_requirements

    # Copied requirements mirror the originals but are distinct objects.
    assert real_openid is not copied_openid
    assert real_foo is not copied_foo
    assert str(real_openid) == str(copied_openid)
    assert str(real_foo) == str(copied_foo)


def _sorted_auth_scope_str(user_app: UserApp) -> str:
    scope_list = user_app.scope_requirements["auth.globus.org"]
    return " ".join(sorted(str(scope) for scope in scope_list))


def test_user_app_get_authorizer():
    client_id = "mock_client_id"
    memory_storage = MemoryTokenStorage()
    memory_storage.store_token_data_by_resource_server(_mock_token_data_by_rs())
    config = GlobusAppConfig(token_storage=memory_storage)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    authorizer = user_app.get_authorizer("auth.globus.org")
    assert isinstance(authorizer, AccessTokenAuthorizer)
    assert authorizer.access_token == "mock_access_token"


def test_user_app_get_authorizer_clears_cache_when_adding_scope_requirements():
    client_id = "mock_client_id"
    memory_storage = MemoryTokenStorage()
    memory_storage.store_token_data_by_resource_server(_mock_token_data_by_rs())
    config = GlobusAppConfig(token_storage=memory_storage)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    initial_authorizer = user_app.get_authorizer("auth.globus.org")
    assert isinstance(initial_authorizer, AccessTokenAuthorizer)
    assert initial_authorizer.access_token == "mock_access_token"

    # We should've cached the authorizer from the first call
    assert user_app.get_authorizer("auth.globus.org") is initial_authorizer

    user_app.add_scope_requirements({"auth.globus.org": [Scope("openid")]})

    # The cache should've been cleared
    updated_authorizer = user_app.get_authorizer("auth.globus.org")
    assert initial_authorizer is not updated_authorizer
    assert isinstance(updated_authorizer, AccessTokenAuthorizer)
    assert updated_authorizer.access_token == "mock_access_token"

    assert user_app.get_authorizer("auth.globus.org") is updated_authorizer


def test_user_app_get_authorizer_refresh():
    client_id = "mock_client_id"
    memory_storage = MemoryTokenStorage()
    memory_storage.store_token_data_by_resource_server(_mock_token_data_by_rs())
    config = GlobusAppConfig(token_storage=memory_storage, request_refresh_tokens=True)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    authorizer = user_app.get_authorizer("auth.globus.org")
    assert isinstance(authorizer, RefreshTokenAuthorizer)
    assert authorizer.refresh_token == "mock_refresh_token"


class CustomExitException(Exception):
    pass


class RaisingLoginFlowManagerCounter(LoginFlowManager):
    """
    A login flow manager which increments a public counter and raises an exception on
    each login attempt.
    """

    def __init__(self) -> None:
        super().__init__(mock.Mock(spec=NativeAppAuthClient))
        self.counter = 0

    def run_login_flow(
        self, auth_parameters: GlobusAuthorizationParameters
    ) -> globus_sdk.OAuthTokenResponse:
        self.counter += 1
        raise CustomExitException("mock login attempt")


def test_user_app_expired_token_triggers_login():
    # Set up token data with an expired access token and no refresh token
    client_id = "mock_client_id"
    memory_storage = MemoryTokenStorage()
    token_data = _mock_token_data_by_rs(
        refresh_token=None,
        expiration_delta=-3600,  # Expired by 1 hour
    )
    memory_storage.store_token_data_by_resource_server(token_data)

    login_flow_manager = RaisingLoginFlowManagerCounter()
    config = GlobusAppConfig(
        token_storage=memory_storage, login_flow_manager=login_flow_manager
    )
    user_app = UserApp("test-app", client_id=client_id, config=config)

    with pytest.raises(CustomExitException):
        user_app.get_authorizer("auth.globus.org")

    assert login_flow_manager.counter == 1


def test_client_app_expired_token_is_auto_resolved():
    """
    This test exercises ClientApp token grant behavior.
    ClientApps may request updated tokens outside the normal token authorization flow.
    """
    client_creds = {
        "client_id": "mock_client_id",
        "client_secret": "mock_client_secret",
    }
    meta = load_response("auth.oauth2_client_credentials_tokens").metadata

    memory_storage = MemoryTokenStorage()
    token_data = _mock_token_data_by_rs(
        resource_server=meta["resource_server"],
        scope=meta["scope"],
        refresh_token=None,
        expiration_delta=-3600,  # Expired by 1 hour
    )
    memory_storage.store_token_data_by_resource_server(token_data)

    config = GlobusAppConfig(token_storage=memory_storage)
    client_app = ClientApp("test-app", **client_creds, config=config)

    transfer = TransferClient(app=client_app, app_scopes=[Scope(meta["scope"])])
    load_response(transfer.task_list)

    starting_token = memory_storage.get_token_data(meta["resource_server"]).access_token
    assert starting_token == token_data[meta["resource_server"]].access_token

    transfer.task_list()

    ending_token = memory_storage.get_token_data(meta["resource_server"]).access_token
    assert starting_token != ending_token
    assert ending_token == meta["access_token"]


def test_client_app_get_authorizer():
    client_id = "mock_client_id"
    client_secret = "mock_client_secret"
    memory_storage = MemoryTokenStorage()
    memory_storage.store_token_data_by_resource_server(_mock_token_data_by_rs())
    config = GlobusAppConfig(token_storage=memory_storage)
    client_app = ClientApp(
        "test-app", client_id=client_id, client_secret=client_secret, config=config
    )

    authorizer = client_app.get_authorizer("auth.globus.org")
    assert isinstance(authorizer, ClientCredentialsAuthorizer)
    assert authorizer.confidential_client.client_id == "mock_client_id"


@mock.patch.object(globus_sdk.IDTokenDecoder, "decode", _mock_decode)
def test_user_app_login_logout(monkeypatch, capsys):
    monkeypatch.setattr("builtins.input", _mock_input)
    load_response(NativeAppAuthClient.oauth2_exchange_code_for_tokens, case="openid")
    load_response(NativeAppAuthClient.oauth2_revoke_token)

    client_id = "mock_client_id"
    memory_storage = MemoryTokenStorage()
    config = GlobusAppConfig(token_storage=memory_storage)
    user_app = UserApp("test-app", client_id=client_id, config=config)

    assert memory_storage.get_token_data("auth.globus.org") is None
    assert user_app.login_required() is True

    user_app.login()
    assert memory_storage.get_token_data("auth.globus.org").access_token is not None
    assert user_app.login_required() is False

    user_app.logout()
    assert memory_storage.get_token_data("auth.globus.org") is None
    assert user_app.login_required() is True


@mock.patch.object(globus_sdk.IDTokenDecoder, "decode", _mock_decode)
def test_client_app_login_logout():
    load_response(
        ConfidentialAppAuthClient.oauth2_client_credentials_tokens, case="openid"
    )
    load_response(ConfidentialAppAuthClient.oauth2_revoke_token)

    client_id = "mock_client_id"
    client_secret = "mock_client_secret"
    memory_storage = MemoryTokenStorage()
    config = GlobusAppConfig(token_storage=memory_storage)
    client_app = ClientApp(
        "test-app", client_id=client_id, client_secret=client_secret, config=config
    )

    assert memory_storage.get_token_data("auth.globus.org") is None

    client_app.login()
    assert memory_storage.get_token_data("auth.globus.org").access_token is not None

    client_app.logout()
    assert memory_storage.get_token_data("auth.globus.org") is None


@mock.patch.object(globus_sdk.IDTokenDecoder, "decode", _mock_decode)
@pytest.mark.parametrize(
    "login_kwargs,expected_login",
    (
        # No params - no additional login
        ({}, False),
        # "force" or "auth_params" - additional login
        ({"force": True}, True),
        (
            {"auth_params": GlobusAuthorizationParameters(session_required_mfa=True)},
            True,
        ),
    ),
)
def test_app_login_flows_can_be_forced(login_kwargs, expected_login, monkeypatch):
    monkeypatch.setattr("builtins.input", _mock_input)
    load_response(NativeAppAuthClient.oauth2_exchange_code_for_tokens, case="openid")

    config = GlobusAppConfig(
        token_storage="memory",
        login_flow_manager=CountingCommandLineLoginFlowManager,
    )
    user_app = UserApp("test-app", client_id="mock_client_id", config=config)

    user_app.login()
    assert user_app.login_required() is False
    assert user_app._login_flow_manager.counter == 1

    user_app.login(**login_kwargs)
    expected_count = 2 if expected_login else 1
    assert user_app._login_flow_manager.counter == expected_count


class CountingCommandLineLoginFlowManager(CommandLineLoginFlowManager):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.counter = 0

    def run_login_flow(
        self,
        auth_parameters: GlobusAuthorizationParameters,
    ) -> globus_sdk.OAuthTokenResponse:
        self.counter += 1
        return super().run_login_flow(auth_parameters)


@pytest.mark.parametrize("config", (None, GlobusAppConfig(token_storage="memory")))
def test_closing_app_closes_implicitly_created_token_storage(config):
    if config is None:
        user_app = UserApp("test-app", client_id="mock_client_id")
    else:
        user_app = UserApp("test-app", client_id="mock_client_id", config=config)

    with mock.patch.object(user_app.token_storage, "close") as token_storage_close:
        user_app.close()
        token_storage_close.assert_called_once()

    user_app.token_storage.close()  # cleanup


def test_closing_app_closes_implicitly_created_login_client():
    user_app = UserApp("test-app", client_id="mock_client_id")

    with mock.patch.object(user_app._login_client, "close") as login_client_close:
        user_app.close()
        login_client_close.assert_called_once()

    user_app._login_client.close()  # cleanup


def test_closing_app_does_not_close_explicitly_passed_token_storage():
    user_app = UserApp(
        "test-app",
        client_id="mock_client_id",
        config=GlobusAppConfig(token_storage=MemoryTokenStorage()),
    )

    with mock.patch.object(user_app.token_storage, "close") as token_storage_close:
        user_app.close()
        token_storage_close.assert_not_called()


def test_app_context_manager_exit_calls_close():
    with mock.patch.object(UserApp, "close") as app_close_method:
        with UserApp("test-app", client_id="mock_client_id"):
            app_close_method.assert_not_called()
        app_close_method.assert_called_once()


def test_app_close_debug_logs_closure_of_resources(caplog):
    """Debug logs show both of the internal clients, plus the token storage"""
    caplog.set_level(logging.DEBUG)

    user_app = UserApp("test-app", client_id="mock_client_id")
    user_app.close()

    # 3 resources at least: consent_client, login_client, token_storage
    assert (
        "closing resource of type AuthClient for UserApp(app_name='test-app')"
        in caplog.text
    )
    assert (
        "closing resource of type NativeAppAuthClient for "
        "UserApp(app_name='test-app')"
    ) in caplog.text
    assert (
        "closing resource of type ValidatingTokenStorage for "
        "UserApp(app_name='test-app')"
    ) in caplog.text



import typing as t

import globus_sdk

specific_flow_client = globus_sdk.SpecificFlowClient("foo")

# running a flow without a notification policy is allowed by the types
specific_flow_client.run_flow({})

# passing a dict[str, t.Any] is allowed (even though the type cannot guarantee safety)
policy_dict: dict[str, t.Any] = {"status": ["INACTIVE"]}
specific_flow_client.run_flow({}, activity_notification_policy=policy_dict)

# passing the object representation is allowed as well
policy_object = globus_sdk.RunActivityNotificationPolicy(status=["FAILED"])
specific_flow_client.run_flow({}, activity_notification_policy=policy_object)

# the object representation does not allow bad values in the status field
globus_sdk.RunActivityNotificationPolicy(status=["FAILURE"])  # type: ignore[list-item]

# passing something of the wrong type (e.g. the status list) is also not allowed
specific_flow_client.run_flow(
    {}, activity_notification_policy=["SUCCEEDED"]  # type: ignore[arg-type]
)



import datetime

import pytest

from globus_sdk import (
    OnceTimerSchedule,
    RecurringTimerSchedule,
    TransferData,
    TransferTimer,
)
from globus_sdk._missing import filter_missing
from tests.common import GO_EP1_ID, GO_EP2_ID


def test_transfer_timer_ok():
    tdata = TransferData(GO_EP1_ID, GO_EP2_ID)
    timer = TransferTimer(body=tdata, name="foo timer", schedule={"type": "once"})
    assert timer["name"] == "foo timer"
    assert timer["schedule"]["type"] == "once"
    assert {"source_endpoint", "destination_endpoint"} < timer["body"].keys()
    assert timer["body"]["source_endpoint"] == GO_EP1_ID
    assert timer["body"]["destination_endpoint"] == GO_EP2_ID


def test_transfer_timer_removes_disallowed_fields():
    # set transfer-data-like body with disallowed fields
    tdata = {"submission_id": "foo", "skip_activation_check": False, "foo": "bar"}
    timer = TransferTimer(body=tdata, name="foo timer", schedule={"type": "once"})
    assert timer["name"] == "foo timer"
    assert timer["schedule"]["type"] == "once"

    # confirm that disallowed fields are stripped in the timer,
    # but the original dict is unchanged
    assert timer["body"] == {"foo": "bar"}
    assert set(tdata.keys()) == {"submission_id", "skip_activation_check", "foo"}


@pytest.mark.parametrize(
    "input_time, expected",
    (
        # even though this string is "obviously" not a valid datetime, we don't
        # translate it when we create the schedule
        ("tomorrow", "tomorrow"),
        # use a fixed (known) timestamp and check how it's formatted as UTC
        (datetime.datetime.fromtimestamp(1698385129.7044), "2023-10-27T05:38:49+00:00"),
        # use a non-UTC datetime and confirm that it is sent as non-UTC
        (
            datetime.datetime.fromisoformat("2023-10-27T05:38:49.999+01:00"),
            "2023-10-27T05:38:49+01:00",
        ),
    ),
)
def test_once_timer_schedule_formats_datetime(input_time, expected):
    schedule = OnceTimerSchedule(datetime=input_time)
    assert dict(schedule) == {"type": "once", "datetime": expected}


def test_recurring_timer_schedule_interval_only():
    schedule = RecurringTimerSchedule(interval_seconds=600)
    assert filter_missing(schedule) == {
        "type": "recurring",
        "interval_seconds": 600,
    }


@pytest.mark.parametrize(
    "input_time, expected",
    (
        # even though this string is "obviously" not a valid datetime, we don't
        # translate it when we create the schedule
        ("tomorrow", "tomorrow"),
        # use a fixed (known) timestamp and check how it's formatted
        (datetime.datetime.fromtimestamp(1698385129.7044), "2023-10-27T05:38:49+00:00"),
    ),
)
def test_recurring_timer_schedule_formats_start(input_time, expected):
    schedule = RecurringTimerSchedule(interval_seconds=600, start=input_time)
    assert filter_missing(schedule) == {
        "type": "recurring",
        "interval_seconds": 600,
        "start": expected,
    }


def test_recurring_timer_schedule_formats_datetime_for_end():
    # use a fixed (known) timestamp and check how it's formatted
    end_time = datetime.datetime.fromtimestamp(1698385129.7044)
    schedule = RecurringTimerSchedule(
        interval_seconds=600, end={"condition": "time", "datetime": end_time}
    )
    assert filter_missing(schedule) == {
        "type": "recurring",
        "interval_seconds": 600,
        "end": {"condition": "time", "datetime": "2023-10-27T05:38:49+00:00"},
    }



from globus_sdk import GCSRoleDocument


def test_gcs_role_helper_supports_additional_fields():
    r = GCSRoleDocument()
    assert r["DATA_TYPE"] == "role#1.0.0"
    assert "foo" not in r

    r2 = GCSRoleDocument(additional_fields={"foo": "bar"})
    assert r2["DATA_TYPE"] == "role#1.0.0"
    assert r2["foo"] == "bar"



import json

import pytest

from globus_sdk import (
    MISSING,
    GroupMemberVisibility,
    GroupPolicies,
    GroupRequiredSignupFields,
    GroupVisibility,
)
from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize(
    "group_vis, group_member_vis, signup_fields, signup_fields_str",
    (
        (
            GroupVisibility.private,
            GroupMemberVisibility.members,
            [GroupRequiredSignupFields.address1],
            ["address1"],
        ),
        (
            GroupVisibility.authenticated,
            GroupMemberVisibility.managers,
            ["address1"],
            ["address1"],
        ),
        (
            "private",
            "members",
            [GroupRequiredSignupFields.address1, "address2"],
            ["address1", "address2"],
        ),
        ("authenticated", "managers", ["address1"], ["address1"]),
    ),
)
def test_set_group_policies(
    groups_manager,
    groups_client,
    group_vis,
    group_member_vis,
    signup_fields,
    signup_fields_str,
):
    group_vis_str = group_vis if isinstance(group_vis, str) else group_vis.value
    group_member_vis_str = (
        group_member_vis
        if isinstance(group_member_vis, str)
        else group_member_vis.value
    )

    meta = load_response(groups_client.set_group_policies).metadata

    resp = groups_manager.set_group_policies(
        meta["group_id"],
        is_high_assurance=False,
        group_visibility=group_vis,
        group_members_visibility=group_member_vis,
        join_requests=False,
        signup_fields=signup_fields,
        authentication_assurance_timeout=28800,
    )
    assert resp.http_status == 200
    assert "address1" in resp.data["signup_fields"]
    # ensure enums were stringified correctly
    req = get_last_request()
    req_body = json.loads(req.body)
    assert req_body["group_visibility"] == group_vis_str
    assert req_body["group_members_visibility"] == group_member_vis_str
    assert req_body["signup_fields"] == signup_fields_str


@pytest.mark.parametrize(
    "group_vis, group_member_vis, signup_fields, signup_fields_str, auth_timeout",
    (
        (
            GroupVisibility.private,
            GroupMemberVisibility.members,
            [GroupRequiredSignupFields.address1],
            ["address1"],
            28800,
        ),
        (
            GroupVisibility.authenticated,
            GroupMemberVisibility.managers,
            ["address1"],
            ["address1"],
            MISSING,
        ),
        (
            "private",
            "members",
            [GroupRequiredSignupFields.address1, "address2"],
            ["address1", "address2"],
            0,
        ),
        ("authenticated", "managers", ["address1"], ["address1"], None),
    ),
)
@pytest.mark.parametrize("setter_usage", (False, "enum", "str"))
def test_set_group_policies_explicit_payload(
    groups_client,
    group_vis,
    group_member_vis,
    signup_fields,
    signup_fields_str,
    auth_timeout,
    setter_usage,
):
    group_vis_str = group_vis if isinstance(group_vis, str) else group_vis.value
    group_member_vis_str = (
        group_member_vis
        if isinstance(group_member_vis, str)
        else group_member_vis.value
    )

    meta = load_response(groups_client.set_group_policies).metadata

    # same payload as the above test, but formulated without GroupsManager
    payload = GroupPolicies(
        is_high_assurance=False,
        group_visibility=group_vis,
        group_members_visibility=group_member_vis,
        join_requests=False,
        signup_fields=signup_fields,
        authentication_assurance_timeout=auth_timeout,
    )
    if setter_usage:
        # set a string in the payload directly
        # this will pass through GroupPolicies.__setitem__
        if setter_usage == "enum":
            payload["group_visibility"] = group_vis
        elif setter_usage == "str":
            payload["group_visibility"] = group_vis_str
        else:
            raise NotImplementedError
    # now send it... (but ignore the response)
    groups_client.set_group_policies(meta["group_id"], payload)
    # ensure enums were stringified correctly, but also that the raw string came through
    req = get_last_request()
    req_body = json.loads(req.body)
    assert req_body["group_visibility"] == group_vis_str
    assert req_body["group_members_visibility"] == group_member_vis_str
    assert req_body["signup_fields"] == signup_fields_str

    # check the authentication_assurance_timeout
    # it should be omitted if it's MISSING
    if auth_timeout is MISSING:
        assert "authentication_assurance_timeout" not in req_body
    else:
        assert req_body["authentication_assurance_timeout"] == auth_timeout



import os

import globus_sdk
from globus_sdk.token_storage import JSONTokenStorage

# this is the tutorial client ID
# replace this string with your ID for production use
CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# define a client for interactions with Globus Auth
auth_client = globus_sdk.NativeAppAuthClient(CLIENT_ID)

# define token storage where tokens will be stored between runs
token_storage = JSONTokenStorage(
    os.path.expanduser("~/.list-my-globus-groups-tokens.json")
)


# if there is no stored token file, we have not yet logged in
if not token_storage.file_exists():
    # do a login flow, getting back a token response
    auth_client.oauth2_start_flow(
        requested_scopes=globus_sdk.GroupsClient.scopes.view_my_groups_and_memberships,
        refresh_tokens=True,
    )
    authorize_url = auth_client.oauth2_get_authorize_url()
    print(f"Please go to this URL and login:\n\n{authorize_url}\n")
    auth_code = input("Please enter the code here: ").strip()
    token_response = auth_client.oauth2_exchange_code_for_tokens(auth_code)
    # now store the tokens
    token_storage.store_token_response(token_response)

# load the tokens from the storage -- either freshly stored or loaded from disk
token_data = token_storage.get_token_data(globus_sdk.GroupsClient.resource_server)

# construct the RefreshTokenAuthorizer which writes back to storage on refresh
authorizer = globus_sdk.RefreshTokenAuthorizer(
    token_data.refresh_token,
    auth_client,
    access_token=token_data.access_token,
    expires_at=token_data.expires_at_seconds,
    on_refresh=token_storage.store_token_response,
)

# use that authorizer to authorize the activity of the groups client
groups_client = globus_sdk.GroupsClient(authorizer=authorizer)

# call out to the Groups service to get a listing
my_groups = groups_client.get_my_groups()

# print in CSV format
print("ID,Name,Roles")
for group in my_groups:
    roles = "|".join({m["role"] for m in group["my_memberships"]})
    print(",".join([group["id"], f'"{group["name"]}"', roles]))



from globus_sdk.authorizers import AccessTokenAuthorizer
from globus_sdk.testing import get_last_request, load_response


def test_get_gcs_info(client):
    meta = load_response(client.get_gcs_info).metadata
    endpoint_client_id = meta["endpoint_client_id"]

    # set an authorizer
    client.authorizer = AccessTokenAuthorizer("access_token")

    res = client.get_gcs_info()
    assert res["endpoint_id"] == endpoint_client_id
    assert res["client_id"] == endpoint_client_id

    # confirm request was unauthenticated despite client having an authorizer
    req = get_last_request()
    assert "Authorization" not in req.headers



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

GUARDIANS_IDP_ID = str(uuid.uuid1())
STAR_LORD = {
    "identity_provider": GUARDIANS_IDP_ID,
    "identity_type": "login",
    "organization": "Guardians of the Galaxy",
    "status": "used",
    "id": str(uuid.uuid1()),
    "name": "Star Lord",
    "username": "star.lord@guardians.galaxy",
    # I thought it would be funny if he didn't get 'star.lord'
    # because someone else got it first
    "email": "star.lord2@guardians.galaxy",
}
ROCKET_RACCOON = {
    "identity_provider": GUARDIANS_IDP_ID,
    "identity_type": "login",
    "organization": "Guardians of the Galaxy",
    "status": "used",
    "id": str(uuid.uuid1()),
    "name": "Rocket",
    "username": "rocket@guardians.galaxy",
    # and think about it, who else would try to lay claim
    # to that email address?
    "email": "star.lord@guardians.galaxy",
}

GUARDIANS_PROJECT = {
    "admin_ids": [ROCKET_RACCOON["id"]],
    "contact_email": "support@guardians.galaxy",
    "display_name": "Guardians of the Galaxy Portal",
    "admin_group_ids": None,
    "id": str(uuid.uuid1()),
    "project_name": "Guardians of the Galaxy Portal",
    "admins": {
        "identities": [STAR_LORD, ROCKET_RACCOON],
        "groups": [],
    },
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path=f"/v2/api/projects/{GUARDIANS_PROJECT['id']}",
        json={"project": GUARDIANS_PROJECT},
        metadata={
            "project_id": GUARDIANS_PROJECT["id"],
        },
    )
)



from globus_sdk.testing import RegisteredResponse, load_response


def test_allow_redirects_false(client):
    # based on a real response from 'GET https://auth.globus.org/'
    load_response(
        RegisteredResponse(
            path="https://foo.api.globus.org/bar",
            status=302,
            headers={
                "Date": "Fri, 15 Apr 2022 15:35:44 GMT",
                "Content-Type": "text/html",
                "Connection": "keep-alive",
                "Server": "nginx",
                "Location": "https://www.globus.org/",
            },
            body="""\
<html>
<head><title>302 Found</title></head>
<body>
<center><h1>302 Found</h1></center>
<hr><center>nginx</center>
</body>
</html>
""",
        )
    )
    # NOTE: this test isn't very "real" because of where `responses` intercepts the
    # request/response action
    # even without `allow_redirects=False`, this test would pass
    # if we find a better way of testing redirect behavior, consider removing this test
    res = client.request("GET", "/bar", allow_redirects=False)
    assert res.http_status == 302


def test_stream_true(client):
    load_response(
        RegisteredResponse(
            path="https://foo.api.globus.org/bar",
            json={"foo": "bar"},
        )
    )
    res = client.request("GET", "/bar", stream=True)
    assert res.http_status == 200
    # forcing JSON evaluation still works as expected (this must force the download /
    # evaluation of content)
    assert res["foo"] == "bar"



import urllib.parse

from globus_sdk.scopes import Scope
from globus_sdk.testing import get_last_request, load_response


def test_oauth2_client_credentials_tokens(auth_client):
    meta = load_response(auth_client.oauth2_client_credentials_tokens).metadata

    response = auth_client.oauth2_client_credentials_tokens(meta["scope"])
    assert (
        response.by_resource_server[meta["resource_server"]]["access_token"]
        == meta["access_token"]
    )


def test_oauth2_client_credentials_tokens_can_accept_scope_object(auth_client):
    meta = load_response(auth_client.oauth2_client_credentials_tokens).metadata

    response = auth_client.oauth2_client_credentials_tokens(Scope(meta["scope"]))
    assert (
        response.by_resource_server[meta["resource_server"]]["access_token"]
        == meta["access_token"]
    )

    last_req = get_last_request()
    assert last_req.body
    body = last_req.body
    assert body != ""
    parsed_body = urllib.parse.parse_qs(body)
    assert parsed_body["scope"] == [meta["scope"]]



from globus_sdk.testing import get_last_request, load_response


def test_list_tunnels(client):
    load_response(client.list_tunnels)

    res = client.list_tunnels()
    assert res.http_status == 200
    assert len(res["data"]) == 2

    req = get_last_request()
    assert req.body is None



from __future__ import annotations

import logging
import typing as t

import globus_sdk
from globus_sdk.scopes import Scope, ScopeParser

from .renewing import RenewingAuthorizer

log = logging.getLogger(__name__)


class ClientCredentialsAuthorizer(
    RenewingAuthorizer["globus_sdk.OAuthClientCredentialsResponse"]
):
    r"""
    Implementation of a RenewingAuthorizer that renews confidential app client
    Access Tokens using a ConfidentialAppAuthClient and a set of scopes to
    fetch a new Access Token when the old one expires.

    Example usage looks something like this:

    >>> import globus_sdk
    >>> confidential_client = globus_sdk.ConfidentialAppAuthClient(
        client_id=..., client_secret=...)
    >>> scopes = "..."
    >>> cc_authorizer = globus_sdk.ClientCredentialsAuthorizer(
    >>>     confidential_client, scopes)
    >>> # create a new client
    >>> transfer_client = globus_sdk.TransferClient(authorizer=cc_authorizer)

    any client that inherits from :class:`BaseClient <globus_sdk.BaseClient>`
    should be able to use a ClientCredentialsAuthorizer to act as
    the client itself.

    :param confidential_client: client object with a valid id and client secret
    :param scopes: A string of space-separated scope names being requested for the
        access tokens that will be used for the Authorization header. These scopes must
        all be for the same resource server, or else the token response will have
        multiple access tokens.
    :param access_token: Initial Access Token to use, only used if ``expires_at`` is
        also set. Must be requested with the same set of scopes passed to this
        authorizer.
    :param expires_at: Expiration time for the starting ``access_token`` expressed as a
        POSIX timestamp (i.e. seconds since the epoch)
    :param on_refresh: A callback which is triggered any time this authorizer fetches a
        new access_token. The ``on_refresh`` callable is invoked on the
        :class:`globus_sdk.OAuthClientCredentialsResponse` object resulting from the
        token being refreshed. It should take only one positional argument, the token
        response object.
        This is useful for implementing storage for Access Tokens, as the
        ``on_refresh`` callback can be used to update the Access Tokens and
        their expiration times.
    """

    def __init__(
        self,
        confidential_client: globus_sdk.ConfidentialAppAuthClient,
        scopes: str | Scope | t.Iterable[str | Scope],
        *,
        access_token: str | None = None,
        expires_at: int | None = None,
        on_refresh: (
            None | t.Callable[[globus_sdk.OAuthClientCredentialsResponse], t.Any]
        ) = None,
    ) -> None:
        # values for _get_token_data
        self.confidential_client = confidential_client
        self.scopes = ScopeParser.serialize(scopes)

        log.debug(
            "Setting up ClientCredentialsAuthorizer with confidential_client="
            f"[instance:{id(confidential_client)}] and scopes={self.scopes}"
        )

        super().__init__(access_token, expires_at, on_refresh)

    def _get_token_response(self) -> globus_sdk.OAuthClientCredentialsResponse:
        """
        Make a request for new tokens, using a 'client_credentials' grant.
        """
        return self.confidential_client.oauth2_client_credentials_tokens(
            requested_scopes=self.scopes
        )

    def _extract_token_data(
        self, res: globus_sdk.OAuthClientCredentialsResponse
    ) -> dict[str, t.Any]:
        """
        Get the tokens .by_resource_server,
        Ensure that only one token was gotten, and return that token.
        """
        token_data = res.by_resource_server.values()
        if len(token_data) != 1:
            raise ValueError(
                "Attempting get new access token for client credentials "
                "authorizer didn't return exactly one token. Ensure scopes "
                f"{self.scopes} are for only one resource server."
            )

        return next(iter(token_data))



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

CREDENTIAL = {
    "name": "foo",
    "id": str(uuid.uuid1()),
    "created": "2023-10-21T22:46:15.845937+00:00",
    "client": "7dee4432-0297-4989-ad23-a2b672a52b12",
    "secret": None,
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        method="DELETE",
        path=f"/v2/api/clients/{CREDENTIAL['client']}/credentials/{CREDENTIAL['id']}",
        json={"credential": CREDENTIAL},
        metadata={
            "credential_id": CREDENTIAL["id"],
            "client_id": CREDENTIAL["client"],
        },
    ),
)
